# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Copyright 2023 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Dict, Optional

from tikka.interfaces.adapters.repository.db_repository import DBRepositoryInterface
from tikka.interfaces.adapters.repository.preferences import (
    PreferencesRepositoryInterface,
)

TABLE_NAME = "preferences"


class DBPreferencesRepository(PreferencesRepositoryInterface, DBRepositoryInterface):
    """
    DBPreferencesRepository class
    """

    def get(self, key: str) -> Optional[str]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            PreferencesRepositoryInterface.get.__doc__
        )

        result = self.client.select_one(
            f"SELECT value_ FROM {TABLE_NAME} WHERE key_=?", (key,)
        )
        if result is None:
            return None

        return result[0]

    def set(self, key: str, value: Optional[str]) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            PreferencesRepositoryInterface.get.__doc__
        )

        # update existing key to value
        self.client.update(TABLE_NAME, f"key_='{key}'", value_=value)

        # if update fails create key with value
        self.client.insert(TABLE_NAME, key_=key, value_=value)

    def delete_all(self) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            PreferencesRepositoryInterface.delete_all.__doc__
        )

        self.client.clear(TABLE_NAME)

    def list(self) -> Dict[str, str]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            PreferencesRepositoryInterface.list.__doc__
        )

        result_set = self.client.select(f"SELECT * FROM {TABLE_NAME}")
        preferences = {}
        for row in result_set:
            preferences[row[0]] = row[1]

        return preferences



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
from typing import Optional

from PyQt5 import QtCore, QtWidgets
from PyQt5.QtCore import QMutex
from PyQt5.QtGui import QKeyEvent, QMovie
from PyQt5.QtWidgets import QApplication, QDialog, QWidget

from tikka.adapters.network.node.node import NetworkNode
from tikka.domains.application import Application
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.events import NodesEvent
from tikka.domains.entities.node import Node
from tikka.slots.pyqt.entities.constants import ICON_LOADER
from tikka.slots.pyqt.entities.worker import AsyncQWorker
from tikka.slots.pyqt.resources.gui.windows.node_add_rc import Ui_nodeAddDialog


class NodeAddWindow(QDialog, Ui_nodeAddDialog):
    """
    NodeAddWindow class
    """

    def __init__(
        self, application: Application, mutex: QMutex, parent: Optional[QWidget] = None
    ):
        """
        Init add node window

        :param application: Application instance
        :param mutex: QMutex instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext

        self.mutex = mutex

        self.network_node_adapter: Optional[NetworkNode] = None
        self.node: Optional[Node] = None
        self.node_prefix: Optional[int] = None
        self.node_genesis_block_hash: Optional[str] = None

        # buttons
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

        # animated loading icon
        self.loader_movie = QMovie(ICON_LOADER)
        self.loader_movie.start()
        self.loaderIconLabel.setMovie(self.loader_movie)
        loader_icon_size_policy = self.loaderIconLabel.sizePolicy()
        loader_icon_size_policy.setRetainSizeWhenHidden(True)
        self.loaderIconLabel.setSizePolicy(loader_icon_size_policy)
        self.loaderIconLabel.hide()

        # events
        self.urlLineEdit.keyPressEvent = self.on_url_line_edit_key_press_event
        self.testButton.clicked.connect(self._on_test_button_clicked)
        self.buttonBox.accepted.connect(self.on_accepted_button)
        self.buttonBox.rejected.connect(self.close)

        ##############################
        # ASYNC METHODS
        ##############################
        # fetch recipient balance
        self.network_get_node_infos_async_qworker = AsyncQWorker(
            self.network_get_node_infos, self.mutex, self
        )
        self.network_get_node_infos_async_qworker.finished.connect(
            self._on_finished_network_get_node_infos
        )

    def on_url_line_edit_key_press_event(self, event: QKeyEvent):
        """
        Triggered when enter is pressed to validate url in url t line edit

        :param event: QKeyEvent instance
        :return:
        """
        if event.key() == QtCore.Qt.Key_Return:
            self._on_test_button_clicked()
        else:
            QtWidgets.QLineEdit.keyPressEvent(self.urlLineEdit, event)
            # if the key is not return, handle normally

    def _on_test_button_clicked(self):
        """
        Run when use click test button

        :return:
        """
        self.loaderIconLabel.show()
        self.testButton.setDisabled(True)
        self.network_get_node_infos_async_qworker.start()

    def network_get_node_infos(self):
        """
        Fetch node infos from the network

        :return:
        """
        url = self.urlLineEdit.text()
        self.network_node_adapter = self.application.nodes.network_get_node_adapter(url)
        if self.network_node_adapter is not None:
            self.node = self.network_node_adapter.get()
            self.node_prefix = self.network_node_adapter.get_ss58_prefix()
            self.node_genesis_block_hash = (
                self.network_node_adapter.get_genesis_block_hash()
            )
        else:
            self.node = None
            self.node_prefix = None
            self.node_genesis_block_hash = None

    def _on_finished_network_get_node_infos(self):
        """
        WHen async network_get_node_infos is finished

        :return:
        """
        self.testButton.setDisabled(False)
        self.loaderIconLabel.hide()

        if self.node is None:
            self.softwareValueLabel.setText("")
            self.versionValueLabel.setText("")
            self.peerIDValueLabel.setText("")
            self.blockValueLabel.setText("")
            self.currencyHashValueLabel.setText("")
            self.currencyPrefixValueLabel.setText("")
            self.errorLabel.setText(self._("Impossible to connect"))
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
        else:
            self.errorLabel.setText("")
            self.softwareValueLabel.setText(self.node.software)
            self.versionValueLabel.setText(self.node.software_version)
            self.peerIDValueLabel.setText(self.node.peer_id)
            self.blockValueLabel.setText(str(self.node.block))
            self.currencyPrefixValueLabel.setText(str(self.node_prefix))
            self.currencyHashValueLabel.setText(self.node_genesis_block_hash)
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(True)

            if (
                self.node_genesis_block_hash
                != self.application.currencies.get_current().genesis_hash
            ):
                self.errorLabel.setText(
                    self._("Current genesis hash is\n{hash}").format(
                        hash=self.application.currencies.get_current().genesis_hash
                    )
                )
                self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            elif (
                self.node_prefix
                != self.application.currencies.get_current().ss58_format
            ):
                self.errorLabel.setText(
                    self._("Current currency prefix is {currency_prefix}").format(
                        currency_prefix=self.application.currencies.get_current().ss58_format
                    )
                )
                self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

    def on_accepted_button(self) -> None:
        """
        Triggered when user click on ok button

        :return:
        """
        url = self.urlLineEdit.text()
        if self.application.nodes.get(url) is not None:
            return

        self.application.nodes.add(
            Node(
                url=self.urlLineEdit.text(),
                peer_id=self.peerIDValueLabel.text(),
                block=int(self.blockValueLabel.text()),
                software=self.softwareValueLabel.text(),
                software_version=self.versionValueLabel.text(),
            )
        )

        self.application.event_dispatcher.dispatch_event(
            NodesEvent(NodesEvent.EVENT_TYPE_LIST_CHANGED)
        )


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    NodeAddWindow(application_, QMutex()).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Optional

from PyQt5.QtCore import QMutex, Qt, QTimer
from PyQt5.QtGui import QFont, QMovie, QPixmap
from PyQt5.QtWidgets import QApplication, QDialog, QWidget
from scalecodec.utils.ss58 import is_valid_ss58_address

from tikka.domains.application import Application
from tikka.domains.entities.account import Account, AccountBalance, AccountCryptoType
from tikka.domains.entities.address import DisplayAddress
from tikka.domains.entities.constants import AMOUNT_UNIT_KEY, DATA_PATH
from tikka.domains.entities.identity import Identity, IdentityStatus
from tikka.domains.entities.smith import Smith, SmithStatus
from tikka.libs import crypto_type
from tikka.libs.keypair import Keypair
from tikka.slots.pyqt.entities.constants import (
    ADDRESS_MONOSPACE_FONT_NAME,
    DEBOUNCE_TIME,
    ICON_IDENTITY_MEMBER,
    ICON_IDENTITY_MEMBER_NOT_VALIDATED,
    ICON_IDENTITY_NOT_MEMBER,
    ICON_LOADER,
    ICON_SMITH,
    SELECTED_UNIT_PREFERENCES_KEY,
)
from tikka.slots.pyqt.entities.worker import AsyncQWorker
from tikka.slots.pyqt.resources.gui.windows.address_add_rc import Ui_AddressAddDialog


class AddressAddWindow(QDialog, Ui_AddressAddDialog):
    """
    AddressAddWindow class
    """

    display_crypto_type = {
        AccountCryptoType.ED25519: "ED25519",
        AccountCryptoType.SR25519: "SR25519",
    }

    def __init__(
        self, application: Application, mutex: QMutex, parent: Optional[QWidget] = None
    ):
        """
        Init add address window

        :param application: Application instance
        :param mutex: QMutex instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext
        self.mutex = mutex

        self.crypto_type = AccountCryptoType.ED25519

        self.address: Optional[str] = None
        self.balance: Optional[AccountBalance] = None
        self.identity: Optional[Identity] = None
        self.smith: Optional[Smith] = None

        self.display_identity_icon = {
            IdentityStatus.UNCONFIRMED.value: QPixmap(
                ICON_IDENTITY_MEMBER_NOT_VALIDATED
            ),
            IdentityStatus.UNVALIDATED.value: QPixmap(
                ICON_IDENTITY_MEMBER_NOT_VALIDATED
            ),
            IdentityStatus.MEMBER.value: QPixmap(ICON_IDENTITY_MEMBER),
            IdentityStatus.NOT_MEMBER.value: QPixmap(ICON_IDENTITY_NOT_MEMBER),
            IdentityStatus.REVOKED.value: QPixmap(ICON_IDENTITY_NOT_MEMBER),
        }

        # calculate balance display by unit preference
        unit_preference = self.application.repository.preferences.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            self.amount = self.application.amounts.get_amount(unit_preference)
        else:
            self.amount = self.application.amounts.get_amount(AMOUNT_UNIT_KEY)

        # set monospace font to address field
        monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        monospace_font.setStyleHint(QFont.Monospace)
        self.addressLineEdit.setFont(monospace_font)

        # animated loading icon
        self.loader_movie = QMovie(ICON_LOADER)
        self.loader_movie.start()
        self.loaderIconLabel.setMovie(self.loader_movie)
        loader_icon_size_policy = self.loaderIconLabel.sizePolicy()
        loader_icon_size_policy.setRetainSizeWhenHidden(True)
        self.loaderIconLabel.setSizePolicy(loader_icon_size_policy)
        self.loaderIconLabel.hide()

        # buttons
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

        # events
        self.buttonBox.accepted.connect(self.on_accepted_button)
        self.addressLineEdit.textChanged.connect(self.on_address_line_edit_changed)
        self.buttonBox.rejected.connect(self.close)

        # debounce timers
        self.address_debounce_timer = QTimer()
        self.address_debounce_timer.timeout.connect(self._address_debounce_call)

        # Create a QWorker object
        self.network_fetch_account_data_async_qworker = AsyncQWorker(
            self.network_fetch_account_data, self.mutex
        )
        self.network_fetch_account_data_async_qworker.finished.connect(
            self._on_finished_network_fetch_account_data
        )

    def on_address_line_edit_changed(self) -> None:
        """
        Triggered when address line edit is changed

        :return:
        """
        if self.address_debounce_timer.isActive():
            self.address_debounce_timer.stop()
        self.address_debounce_timer.start(DEBOUNCE_TIME)

    def _address_debounce_call(self):
        """
        Debounce function triggered only after idle time when typing address

        :return:
        """
        # stop mnemonic_debounce_timer to avoid infinite loop
        if self.address_debounce_timer.isActive():
            self.address_debounce_timer.stop()

        self._check_address_validity()

        if self.address and self.application.connections.node.is_connected():
            self.loaderIconLabel.show()
            self.network_fetch_account_data_async_qworker.start()

    def _check_address_validity(self) -> Optional[DisplayAddress]:
        """
        Validate address and store it if valid

        :return:
        """

        self.errorLabel.setText("")
        self.address = None
        address = self.addressLineEdit.text().strip()
        if not is_valid_ss58_address(address):
            self.errorLabel.setText(self._("Account address is not valid!"))
            self.keyTypeValueLabel.setText("?")
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            return None

        if self.application.accounts.get_by_address(address) is not None:
            self.errorLabel.setText(self._("Account already exists!"))
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            return None

        keypair = Keypair(address)
        try:
            result = crypto_type.is_valid_ed25519(keypair.public_key)
        except AttributeError:
            result = False
        except AssertionError:
            result = False

        if result is True:
            self.crypto_type = AccountCryptoType.ED25519
            self.keyTypeValueLabel.setText(
                self.display_crypto_type[AccountCryptoType.ED25519]
            )
        else:
            try:
                result = crypto_type.is_valid_sr25519(keypair.public_key)
            except AttributeError:
                result = False
            except AssertionError:
                result = False
            if result is True:
                self.crypto_type = AccountCryptoType.SR25519
                self.keyTypeValueLabel.setText(
                    self.display_crypto_type[AccountCryptoType.SR25519]
                )

        self.address = address
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(True)
        return None

    def network_fetch_account_data(self):
        """
        Fetch account data and
        check if account is really a known legacy v1 account

        :return:
        """
        address = self.address
        self.balance = None
        self.identity = None
        self.smith = None

        try:
            self.balance = self.application.accounts.network_get_balance(address)
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)
        else:
            try:
                self.identity = self.application.identities.network_get_identity(
                    address
                )
            except Exception as exception:
                self.errorLabel.setText(self._(str(exception)))
                logging.exception(exception)
            else:
                if self.identity is not None:
                    try:
                        self.smith = self.application.smiths.network_get_smith(
                            self.identity.index
                        )
                    except Exception as exception:
                        self.errorLabel.setText(self._(str(exception)))
                        logging.exception(exception)

    def _on_finished_network_fetch_account_data(self):
        """
        Triggered when async request network_fetch_account_data_async_qworker is finished

        :return:
        """
        self.loaderIconLabel.hide()

        if self.balance is not None:
            self.balanceValueLabel.setText(
                self.locale().toCurrencyString(
                    self.amount.value(self.balance.total), self.amount.symbol()
                )
            )
        else:
            self.balanceValueLabel.setText("?")
        if self.identity:
            self.identityIconLabel.setPixmap(
                self.display_identity_icon[self.identity.status.value].scaled(
                    16, 18, aspectRatioMode=Qt.KeepAspectRatio
                )
            )
            identity_name = self.identity.name or ""
            self.identityNameAndIndexValuelabel.setText(
                f"{identity_name}#{self.identity.index}"
            )
            if self.smith is not None and self.smith.status == SmithStatus.SMITH:
                self.identityIconLabel.setPixmap(
                    QPixmap(ICON_SMITH).scaled(
                        16, 18, aspectRatioMode=Qt.KeepAspectRatio
                    )
                )
        else:
            self.identityNameAndIndexValuelabel.setText("")
            self.identityIconLabel.clear()

    def on_accepted_button(self) -> None:
        """
        Triggered when user click on ok button

        :return:
        """
        if self.address is not None:
            name = self.nameLineEdit.text().strip()

            # create account instance
            account = Account(
                address=self.address, name=name, crypto_type=self.crypto_type
            )

            if self.balance:
                account.balance = self.balance.total
                account.balance_available = self.balance.available
                account.balance_reserved = self.balance.reserved

            self.application.accounts.add(account)

            if self.identity:
                if self.application.identities.get(self.identity.index) is None:
                    self.application.identities.add(self.identity)
                else:
                    self.application.identities.update(self.identity)

                if self.smith:
                    if self.application.smiths.get(self.identity.index) is None:
                        self.application.smiths.add(self.smith)
                    else:
                        self.application.smiths.update(self.smith)


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    AddressAddWindow(application_, QMutex()).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Any, Dict, List, Optional
from uuid import UUID

from tikka.domains.currencies import Currencies
from tikka.domains.entities.account import Account, AccountBalance, AccountCryptoType
from tikka.domains.entities.constants import DERIVATION_SCAN_MAX_NUMBER
from tikka.domains.entities.events import AccountEvent, LastTransferEvent
from tikka.domains.events import EventDispatcher
from tikka.domains.passwords import Passwords
from tikka.domains.transfers import Transfers
from tikka.domains.wallets import Wallets
from tikka.interfaces.adapters.network.indexer.accounts import IndexerAccountsInterface
from tikka.interfaces.adapters.network.node.accounts import NodeAccountsInterface
from tikka.interfaces.adapters.repository.accounts import AccountsRepositoryInterface
from tikka.interfaces.adapters.repository.file_wallets import (
    V1FileWalletsRepositoryInterface,
)
from tikka.libs.derivation import RE_TRANSPARENT_DERIVATION_PATH_PATTERN
from tikka.libs.keypair import Keypair
from tikka.libs.signing_key_v1 import SigningKey


class Accounts:
    """
    Accounts domain class
    """

    def __init__(
        self,
        repository: AccountsRepositoryInterface,
        network_node: NodeAccountsInterface,
        network_indexer: IndexerAccountsInterface,
        passwords: Passwords,
        wallets: Wallets,
        transfers: Transfers,
        file_wallets_repository: V1FileWalletsRepositoryInterface,
        currencies: Currencies,
        event_dispatcher: EventDispatcher,
    ):
        """
        Init Accounts domain

        :param repository: AccountsRepositoryInterface instance
        :param network_node: NodeAccountsInterface instance
        :param network_indexer: IndexerAccountsInterface instance
        :param passwords: Passwords domain instance
        :param wallets: Wallets domain instance
        :param transfers: Transfers domain instance
        :param file_wallets_repository: FileWalletsRepository adapter instance
        :param currencies: Currencies domain instance
        :param event_dispatcher: EventDispatcher instance
        """
        self.repository = repository
        self.network_node = network_node
        self.network_indexer = network_indexer
        self.passwords = passwords
        self.wallets = wallets
        self.transfers = transfers
        self.file_wallets_repository = file_wallets_repository
        self.currencies = currencies
        self.event_dispatcher = event_dispatcher

        # subscribe to events
        self.event_dispatcher.add_event_listener(
            LastTransferEvent.EVENT_TYPE_LAST_TRANSFER_CHANGED,
            self.on_last_transfer_changed,
        )

    @staticmethod
    def get_instance(address: str, name: Optional[str] = None) -> Account:
        """
        Return an Account instance from arguments

        :param address: Account address
        :param name: Optional account name
        :return:
        """
        return Account(address=address, name=name)

    def get_list(
        self,
        filters: Optional[Dict[str, Any]] = None,
        sort_column: Optional[str] = None,
        sort_order: str = AccountsRepositoryInterface.SORT_ORDER_ASCENDING,
    ):
        """
        Return accounts from repository with optional filters and sort_column

        :param filters: Dict with {column: value} filters or None
        :param sort_column: Sort column constant like COLUMN_ADDRESS or None
        :param sort_order: Sort order constant SORT_ORDER_ASCENDING or SORT_ORDER_DESCENDING
        :return:
        """
        # get accounts from database
        return self.repository.list(filters, sort_column, sort_order)

    def add(self, account: Account):
        """
        Add account

        :param account: Account instance
        :return:
        """
        # add account
        self.repository.add(account)

        # dispatch event
        event = AccountEvent(
            AccountEvent.EVENT_TYPE_ADD,
            account,
        )
        self.event_dispatcher.dispatch_event(event)

    def update(self, account: Account):
        """
        Update account

        :param account: Account instance
        :return:
        """
        self.repository.update(account)

        # dispatch event
        event = AccountEvent(
            AccountEvent.EVENT_TYPE_UPDATE,
            account,
        )
        self.event_dispatcher.dispatch_event(event)

    def get_by_index(self, index: int) -> Account:
        """
        Return account instance from index

        :param index: Index in account list
        :return:
        """
        return self.get_list()[index]

    def get_by_address(self, address: str) -> Optional[Account]:
        """
        Return account instance from address

        :param address: Account address
        :return:
        """
        for account in self.get_list():
            if account.address == address:
                return account

        return None

    def delete(self, account: Account) -> None:
        """
        Delete account in list and repository

        :param account: Account instance to delete
        :return:
        """
        self.repository.delete(account.address)

        # dispatch event
        event = AccountEvent(
            AccountEvent.EVENT_TYPE_DELETE,
            account,
        )
        self.event_dispatcher.dispatch_event(event)

    def delete_all(self) -> None:
        """
        Delete all accounts in repository

        :return:
        """
        self.repository.delete_all()

    def count(self) -> int:
        """
        Return total number of accounts

        :return:
        """
        return self.repository.count()

    def get_total_balance(self) -> int:
        """
        Return total balance of all accounts

        :return:
        """
        return self.repository.total_balance()

    def unlock(self, account: Account, wallet_password: str) -> bool:
        """
        Unlock wallet of account if wallet_password is OK

        :param account: Account instance
        :param wallet_password: Passphrase
        :return:
        """
        result = self.wallets.unlock(account.address, wallet_password)
        if result is False:
            return False

        self.update(account)
        return True

    def lock(self, account: Account):
        """
        Lock account

        :param account: Account instance
        :return:
        """
        self.wallets.lock(account.address)

        # dispatch event
        event = AccountEvent(
            AccountEvent.EVENT_TYPE_UPDATE,
            account,
        )
        self.event_dispatcher.dispatch_event(event)

    def forget_wallet(self, account: Account) -> None:
        """
        Delete stored wallet for this account

        :return:
        """
        self.wallets.delete(account.address)
        # update account display
        self.event_dispatcher.dispatch_event(
            AccountEvent(AccountEvent.EVENT_TYPE_UPDATE, account)
        )

    def network_update_balance(self, account: Account) -> Account:
        """
        Update account balance from network

        :param account: Account instance
        :return:
        """
        account_balance = self.network_node.get_balance(account.address)
        if account_balance is not None:
            account.balance = account_balance.total
            account.balance_available = account_balance.available
            account.balance_reserved = account_balance.reserved
        else:
            account.balance = None
            account.balance_available = None
            account.balance_reserved = None
        self.repository.update(account)

        return account

    def network_update_account(self, account: Account) -> Account:
        """
        Update
        - balance
        - total_transfers_count
        from network

        :param account: Account instance
        :return:
        """
        account_balance = self.network_node.get_balance(account.address)
        if account_balance is not None:
            account.balance = account_balance.total
            account.balance_available = account_balance.available
            account.balance_reserved = account_balance.reserved
        else:
            account.balance = None
            account.balance_available = None
            account.balance_reserved = None
        self.repository.update(account)

        return account

    def network_update_balances(self, accounts: List[Account]) -> None:
        """
        Update balances of account list from network

        :param accounts: Account list
        :return:
        """
        balances = self.network_node.get_balances(
            [account.address for account in accounts]
        )
        for account in accounts:
            if balances[account.address] is not None:
                account.balance = balances[account.address].total  # type: ignore
                account.balance_available = balances[account.address].available  # type: ignore
                account.balance_reserved = balances[account.address].reserved  # type: ignore
            else:
                account.balance = None
                account.balance_available = None
                account.balance_reserved = None
            self.repository.update(account)

    def network_get_balance(self, address: str) -> Optional[AccountBalance]:
        """
        Return AccountBalance instance from network

        :param address: Account address
        :return:
        """
        return self.network_node.get_balance(address)

    def network_get_balances(
        self, addresses: List[str]
    ) -> Dict[str, Optional[AccountBalance]]:
        """
        Return balances of address list from network

        :param addresses: Account address list
        :return:
        """
        return self.network_node.get_balances(addresses)

    def network_update_total_transfers_count(self, account: Account) -> Account:
        """
        Update account total_transfers_count from network

        :param account: Account to update
        :return:
        """
        account.total_transfers_count = (
            self.transfers.network_fetch_total_count_for_address(account.address)
        )
        self.repository.update(account)
        return account

    def list_by_category_id(self, category_id: Optional[UUID]) -> List[Account]:
        """
        Return all accounts in category_id

        :param category_id: Category ID
        :return:
        """
        result = []
        for account in self.get_list():
            # do not list derived accounts
            if account.category_id is None and account.root is not None:
                continue
            if account.category_id == category_id:
                result.append(account)

        return result

    def get_derivation_accounts(self, address: str) -> List[Account]:
        """
        Return list of derivation accounts from this root address

        :param address: Root account address
        :return:
        """
        filters = {AccountsRepositoryInterface.COLUMN_ROOT: address}
        # sort ascending by derivation path numbers
        return self.repository.list(
            filters=filters,
            sort_column=AccountsRepositoryInterface.COLUMN_PATH,
            sort_order=AccountsRepositoryInterface.SORT_ORDER_ASCENDING,
        )

    def create_new_account(
        self,
        mnemonic: str,
        language_code: str,
        derivation: str,
        crypto_type: AccountCryptoType,
        name: str,
        password: str,
        add_event: bool = True,
    ) -> Account:
        """
        Create a root account (read-only if derivation not empty) with mnemonic, if it does not already exist,
        and create a derived account if derivation is not "" as per RFC 0019.
        Return root account if derivation == "", derived account otherwise.

        :param mnemonic: Mnemonic phrase
        :param language_code: Mnemonic language code
        :param derivation: Derivation path
        :param crypto_type: Key type as AccountCryptoType instance
        :param name: Account name
        :param password: Wallet password
        :param add_event: Trigger event if True
        :return:
        """
        # get root keypair
        root_keypair = Keypair.create_from_mnemonic(
            mnemonic=mnemonic,
            language_code=language_code,
            crypto_type=crypto_type,
            ss58_format=self.currencies.get_current().ss58_format,
        )
        existing_account = self.get_by_address(root_keypair.ss58_address)
        if existing_account is None:
            # store new password for new root account
            if derivation == "":
                self.create_new_root_account(
                    mnemonic, language_code, crypto_type, name, password
                )
            else:
                # read-only root account
                self.create_new_root_account(
                    mnemonic, language_code, crypto_type, None, None, False
                )
            self.passwords.new(root_keypair, password)
            root_account = self.get_by_address(root_keypair.ss58_address)
        else:
            # get password of root account
            clear_password = self.passwords.get_clear_password(root_keypair)
            if clear_password is not None:
                password = clear_password
            root_account = existing_account

        if derivation == "":
            return root_account  # type: ignore

        # create keypair from mnemonic and path
        keypair = Keypair.create_from_uri(
            suri=mnemonic + derivation,
            language_code=language_code,
            crypto_type=crypto_type,
            ss58_format=self.currencies.get_current().ss58_format,
        )

        # create and store Account instance
        account = Account(
            keypair.ss58_address,
            crypto_type=crypto_type,
            legacy_v1=False,
            name=None if name == "" else name,
            root=root_account.address,  # type: ignore
            path=derivation,
        )

        if add_event:
            self.add(account)
        else:
            self.repository.add(account)

        # create and store Wallet instance
        wallet = self.wallets.create(keypair, password)
        if not self.wallets.exists(keypair.ss58_address):
            self.wallets.add(wallet)
            self.wallets.unlock(keypair.ss58_address, password)

        return account

    def create_new_root_account(
        self,
        mnemonic: str,
        language_code: str,
        crypto_type: AccountCryptoType,
        name: Optional[str],
        password: Optional[str],
        add_event: bool = True,
    ) -> Account:
        """
        Create and return a root account from mnemonic

        :param mnemonic: Mnemonic phrase
        :param language_code: Mnemonic language code
        :param crypto_type: Key type as AccountCryptoType instance
        :param name: Optional account name
        :param password: Optional Wallet password, if None no wallet will be created
        :param add_event: Optional, Default True to trigger add event
        :return:
        """

        # get root keypair
        root_keypair = Keypair.create_from_mnemonic(
            mnemonic=mnemonic,
            language_code=language_code,
            crypto_type=crypto_type,
            ss58_format=self.currencies.get_current().ss58_format,
        )

        root_account = Account(
            root_keypair.ss58_address, crypto_type=crypto_type, name=name
        )

        if add_event:
            self.add(root_account)
        else:
            # add root account in repository without event
            self.repository.add(root_account)

        if password is not None:
            # create and store Wallet instance
            wallet = self.wallets.new(root_keypair, password)
            self.wallets.unlock(wallet.address, password)
            # store new password for new V1 root account
            self.passwords.new(root_keypair, password)

        return root_account

    def create_new_root_account_v1_from_credentials(
        self, secret_id: str, password_id: str, name: str, password: str
    ) -> Optional[Account]:
        """
        Create an unlocked V1 root account with credentials secret_id and password_id

        :param secret_id: Secret ID
        :param password_id: Password ID
        :param name: Name of account
        :param password: Password of wallet
        :return:
        """
        signing_key = SigningKey.from_credentials(secret_id, password_id)
        return self.create_new_root_account_v1_from_seed(
            signing_key.seed.hex(), name, password
        )

    def create_new_root_account_v1_from_seed(
        self, seed_hex: str, name: str, password: str
    ) -> Optional[Account]:
        """
        Create an unlocked V1 root account from hexadecimal seed string

        :param seed_hex: Hexadecimal seed string
        :param name: Name of account
        :param password: Password of wallet
        :return:
        """
        keypair = Keypair.create_from_seed(
            seed_hex=seed_hex,
            ss58_format=self.currencies.get_current().ss58_format,
            crypto_type=AccountCryptoType.ED25519,
        )
        address = keypair.ss58_address
        account = self.get_by_address(address)
        if account is not None:
            return account

        # create and store Account instance
        account = Account(
            address,
            crypto_type=AccountCryptoType.ED25519,
            legacy_v1=True,
            name=None if name == "" else name,
        )
        self.add(account)

        # create and store Wallet instance
        wallet = self.wallets.create(keypair, password)
        self.wallets.add(wallet)
        self.wallets.unlock(wallet.address, password)

        # store new password for new V1 root account
        self.passwords.new(keypair, password)

        return account

    def get_available_derivation_list(self, root_account: Account) -> List[str]:
        """
        Return a list of derivation accounts from root account given, not already in account list

        :param root_account: Root account instance
        :return:
        """
        derived_accounts = self.get_derivation_accounts(root_account.address)
        available_derivation_list = []

        # get max existing transparent derivation number
        derivation_numbers: List[int] = []
        max_derivation_number = 0
        for account in derived_accounts:
            if account.path is None:
                continue
            match = RE_TRANSPARENT_DERIVATION_PATH_PATTERN.fullmatch(account.path)
            if match:
                account_derivation_number = int(match.group(1))
                derivation_numbers.append(account_derivation_number)
                max_derivation_number = max(
                    max_derivation_number, account_derivation_number
                )

        for derivation_number in range(
            0, max(max_derivation_number, DERIVATION_SCAN_MAX_NUMBER) + 1, 1
        ):
            if derivation_number not in derivation_numbers:
                # add next available transparent derivation path
                available_derivation_list.append(f"//{derivation_number}")

        return available_derivation_list

    def update_category_id(
        self, category_id: UUID, new_category_id: Optional[UUID]
    ) -> None:
        """
        Move all accounts with category_id to new_category_id

        :param category_id: Category ID
        :param new_category_id: New category ID
        :return:
        """
        for account in self.list_by_category_id(category_id):
            account.category_id = new_category_id
            self.update(account)

    def reset_password(
        self, root_keypair: Keypair, mnemonic: str, language_code: str, password: str
    ):
        """
        Update wallet encryption with password for all wallets derived from mnemonic

        :param root_keypair: Keypair of root account
        :param mnemonic: Mnemonic phrase
        :param language_code: Language code
        :param password: Password to set
        :return:
        """
        # if root wallet exists...
        if self.wallets.exists(root_keypair.ss58_address):
            # update root wallet with new password
            self.wallets.update(self.wallets.create(root_keypair, password))

        for account in self.get_derivation_accounts(root_keypair.ss58_address):
            # if account wallet exists...
            if self.wallets.exists(account.address) and account.path is not None:
                # update account wallet with new password
                suri = mnemonic + account.path
                keypair = Keypair.create_from_uri(
                    suri=suri,
                    language_code=language_code,
                    crypto_type=root_keypair.crypto_type,
                    ss58_format=root_keypair.ss58_format,
                )
                self.wallets.update(self.wallets.create(keypair, password))

    def on_last_transfer_changed(self, event: LastTransferEvent):
        """
        Update account with last transfer timestamp got from network

        :param event: LastTransferEvent instance
        :return:
        """
        account = self.get_by_address(event.address)
        if account is not None:
            account.last_transfer_timestamp = event.transfer.timestamp
            self.repository.update(account)

    def network_is_legacy_v1(self, address: str) -> bool:
        """
        Return True if account is a legacy v1 account (present in genesis)

        :param address: ss58 account address
        :return:
        """
        return self.network_indexer.is_legacy_v1(address)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
from typing import Optional

from gql import Client, gql
from gql.client import SyncClientSession
from gql.transport.requests import RequestsHTTPTransport

from tikka.interfaces.adapters.network.connection import ConnectionInterface
from tikka.interfaces.entities.Server import ServerInterface

# graphQL timeout
GRAPHQL_CONNECTION_TIMEOUT = 30


class IndexerConnection(ConnectionInterface):
    """
    IndexerConnection class
    """

    def __init__(self) -> None:
        """
        Init IndexerConnection instance

        gql session is available in self.client after connect()
        """
        self.client: Optional[SyncClientSession] = None

    def connect(self, server: ServerInterface) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionInterface.connect.__doc__
        )
        self.url = server.url
        logging.debug(
            "CONNECTING TO INDEXER %s......................................", server.url
        )
        # Create a GraphQL client using the defined transport
        gql_client = Client(
            transport=RequestsHTTPTransport(url=server.url),
        )

        try:
            self.client = gql_client.connect_sync()
            # check connection...
            query = gql(
                """
            query {
                blocks (
                    orderBy: HEIGHT_DESC
                    first: 1
                ) {
                    nodes {
                        height
                    }
                }
            }

            """
            )
            result = self.client.execute(query)  # type: ignore
        except Exception as exception:
            logging.exception(exception)
            self.client = None
            logging.debug(
                "CONNECTION TO INDEXER %s......................................FAILED !",
                server.url,
            )
        else:
            if len(result["blocks"]) < 1:
                self.client = None
                logging.debug(
                    "CONNECTION TO INDEXER %s......................................FAILED !",
                    server.url,
                )
            else:
                logging.debug(
                    "CONNECTION TO INDEXER %s......................................SUCCESS !",
                    server.url,
                )

    def disconnect(self) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionInterface.disconnect.__doc__
        )
        if self.client is not None:
            self.client.close()
            self.client = None
            logging.debug("GraphQL connection closed.")

    def is_connected(self) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionInterface.is_connected.__doc__
        )
        return self.client is not None



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import abc
from typing import TYPE_CHECKING

from tikka.interfaces.adapters.network.connection import ConnectionInterface

if TYPE_CHECKING:
    from tikka.domains.entities.node import Node
    from tikka.interfaces.adapters.network.node.accounts import NodeAccountsInterface
    from tikka.interfaces.adapters.network.node.authorities import (
        NodeAuthoritiesInterface,
    )
    from tikka.interfaces.adapters.network.node.currency import NodeCurrencyInterface
    from tikka.interfaces.adapters.network.node.identities import (
        NodeIdentitiesInterface,
    )
    from tikka.interfaces.adapters.network.node.smiths import NodeSmithsInterface
    from tikka.interfaces.adapters.network.node.technical_committee import (
        NodeTechnicalCommitteeInterface,
    )
    from tikka.interfaces.adapters.network.node.transfers import NodeTransfersInterface


class NetworkNodeInterface(abc.ABC):
    """
    NetworkNodeInterface class
    """

    @abc.abstractmethod
    def get(self) -> Node:
        """
        Return the node instance with infos from connection

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_ss58_prefix(self) -> int:
        """
        Return the node ss58 prefix from connection

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_genesis_block_hash(self) -> str:
        """
        Return the node genesis block hash from connection

        :return:
        """
        raise NotImplementedError

    @property
    def connection(self) -> ConnectionInterface:
        """
        Return ConnectionsInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def accounts(self) -> NodeAccountsInterface:
        """
        Return NodeAccountsInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def authorities(self) -> NodeAuthoritiesInterface:
        """
        Return NodeAuthoritiesInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def currency(self) -> NodeCurrencyInterface:
        """
        Return NodeCurrencyInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def identities(self) -> NodeIdentitiesInterface:
        """
        Return NodeIdentitiesInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def smiths(self) -> NodeSmithsInterface:
        """
        Return NodeSmithsInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def technical_committee(self) -> NodeTechnicalCommitteeInterface:
        """
        Return NodeTechnicalCommitteeInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def transfers(self) -> NodeTransfersInterface:
        """
        Return NodeTransfersInterface instance

        :return:
        """
        raise NotImplementedError


class NetworkNodeException(Exception):
    """
    NetworkNodeException class
    """



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
import logging
import random
from json import JSONDecodeError
from typing import List, Optional
from urllib import request

from tikka.adapters.network.indexer.indexer import NetworkIndexer
from tikka.domains.config import Config
from tikka.domains.currencies import Currencies
from tikka.domains.entities.constants import (
    INDEXERS_CURRENT_ENTRY_POINT_URL_PREFERENCES_KEY,
)
from tikka.domains.entities.events import IndexersEvent
from tikka.domains.entities.indexer import Indexer
from tikka.domains.events import EventDispatcher
from tikka.domains.preferences import Preferences
from tikka.interfaces.adapters.network.indexer.indexer import NetworkIndexerInterface
from tikka.interfaces.adapters.repository.indexers import IndexersRepositoryInterface


class Indexers:
    """
    Indexers domain class
    """

    CONFIG_INDEXERS_ENDPOINTS_KEYWORD = "squid"
    CONFIG_ONLINE_ENDPOINTS_KEYWORD = "online"

    def __init__(
        self,
        repository: IndexersRepositoryInterface,
        preferences: Preferences,
        network_indexer: NetworkIndexerInterface,
        config: Config,
        currencies: Currencies,
        event_dispatcher: EventDispatcher,
    ):
        """
        Init Indexers domain instance

        :param repository: EntryPointsRepositoryInterface instance
        :param preferences: Preferences domain instance
        :param network_indexer: Network adapter instance for handling indexers
        :param config: Config instance
        :param currencies: Currencies instance
        :param event_dispatcher: EventDispatcher instance
        """
        self.repository = repository
        self.preferences = preferences
        self.network_indexer = network_indexer
        self.config = config
        self.currencies = currencies
        self.event_dispatcher = event_dispatcher
        self._current_url = self.currencies.get_entry_point_urls()[
            self.CONFIG_INDEXERS_ENDPOINTS_KEYWORD
        ][0]

        self.init_repository()

    def init_repository(self):
        """
        Init repository with default entry points from config

        :return:
        """
        repository_urls = self.repository.get_urls()
        currency_endpoints = self.currencies.get_entry_point_urls()

        # init repository with current currency entry point urls
        for url in currency_endpoints[self.CONFIG_INDEXERS_ENDPOINTS_KEYWORD]:
            if url not in repository_urls:
                self.repository.add(Indexer(url))

        self._current_url = self.repository.list(0, 1)[0].url

        current_url_in_preferences = self.preferences.get(
            INDEXERS_CURRENT_ENTRY_POINT_URL_PREFERENCES_KEY
        )
        if (
            current_url_in_preferences is None
            or current_url_in_preferences not in self.repository.get_urls()
        ):
            self.preferences.set(
                INDEXERS_CURRENT_ENTRY_POINT_URL_PREFERENCES_KEY, self.get_current_url()
            )
        else:
            self._current_url = current_url_in_preferences

    def network_fetch_endpoints(self):
        """
        Add Indexers in repository from online list of endpoints

        :return:
        """
        currency_endpoints = self.currencies.get_entry_point_urls()

        # fetch endpoints from online json file
        if currency_endpoints[self.CONFIG_ONLINE_ENDPOINTS_KEYWORD]:
            online_url = currency_endpoints[self.CONFIG_ONLINE_ENDPOINTS_KEYWORD]
            try:
                with request.urlopen(online_url) as file:
                    online_endpoints = json.load(file)
            except JSONDecodeError as exception:
                logging.warning("error loading endpoints from %s", online_url)
                logging.exception(exception)
                online_endpoints = None

            if online_endpoints is not None:
                repository_urls = self.repository.get_urls()
                for url in online_endpoints[self.CONFIG_INDEXERS_ENDPOINTS_KEYWORD]:
                    if url not in repository_urls:
                        self.repository.add(Indexer(url))

    def network_fetch_current_indexer(self) -> None:
        """
        Update indexer from network

        :return:
        """
        current_indexer = self.repository.get(self.get_current_url())
        network_indexer = self.network_indexer.get()
        if network_indexer is None:
            return None

        if current_indexer is not None:
            # update only changing properties
            current_indexer.block = network_indexer.block

            self.repository.update(current_indexer)

        return None

    def network_get_genesis_hash(self) -> str:
        """
        Get from network and return indexer genesis hash

        :return:
        """
        return self.network_indexer.get_genesis_hash()

    @staticmethod
    def network_test_and_get_indexer(url: str) -> Optional[Indexer]:
        """
        Try to open connection on url and return indexer if successful

        Then close connection

        :param url: Entry point url
        :return:
        """
        indexer = None

        network_indexer = NetworkIndexer()
        network_indexer.connection.connect(Indexer(url))
        if network_indexer.connection.is_connected():
            indexer = network_indexer.get()
        network_indexer.connection.disconnect()

        return indexer

    def network_set_url_randomly(self):
        """
        Shuffle indexers list randomly and connect to first available indexer

        :return:
        """
        indices = list(range(1, self.count()))
        random.shuffle(indices)
        for index in indices:
            url = self.list()[index].url
            # never choose localhost randomly...
            if "localhost" not in url:
                self.network_indexer.connection.connect(Indexer(url))
                indexer_genesis_hash = self.network_indexer.get_genesis_hash()
                if indexer_genesis_hash != self.currencies.get_current().genesis_hash:
                    self.network_indexer.connection.disconnect()
                if self.network_indexer.connection.is_connected():
                    self.set_current_url(url)
                    break

    def add(self, indexer: Indexer) -> None:
        """
        Add indexer in repository

        :param indexer: Indexer instance
        :return:
        """
        self.repository.add(indexer)

        self.event_dispatcher.dispatch_event(
            IndexersEvent(IndexersEvent.EVENT_TYPE_LIST_CHANGED)
        )

    def get(self, url: str) -> Optional[Indexer]:
        """
        Get Indexer instance by url

        :param url: Url
        :return:
        """
        return self.repository.get(url)

    def update(self, indexer: Indexer) -> None:
        """
        Update indexer in repository

        :param indexer: Indexer instance
        :return:
        """
        self.repository.update(indexer)

        self.event_dispatcher.dispatch_event(
            IndexersEvent(IndexersEvent.EVENT_TYPE_LIST_CHANGED)
        )

    def list(self) -> List[Indexer]:
        """
        Return all Indexers from repository

        :return:
        """
        return self.repository.list()

    def count(self) -> int:
        """
        Return total indexers count

        :return:
        """
        return self.repository.count()

    def delete(self, url: str) -> None:
        """
        Delete Indexer by url

        :param url: Indexer url
        :return:
        """
        # do not delete default entry points from config
        if (
            url
            in self.currencies.get_entry_point_urls()[
                self.CONFIG_INDEXERS_ENDPOINTS_KEYWORD
            ]
        ):
            return

        self.repository.delete(url)
        # switch current entry point to first in list
        self.set_current_url(self.repository.list(0, 1)[0].url)
        # set new entry point in preferences
        self.preferences.set(
            INDEXERS_CURRENT_ENTRY_POINT_URL_PREFERENCES_KEY, self.get_current_url()
        )

        self.event_dispatcher.dispatch_event(
            IndexersEvent(IndexersEvent.EVENT_TYPE_LIST_CHANGED)
        )

    def delete_all(self) -> None:
        """
        Delete all Indexers in repository

        :return:
        """
        self.repository.delete_all()

    def get_current_url(self) -> str:
        """
        Return current entry point url

        :return:
        """
        return self._current_url

    def set_current_url(self, url: str) -> None:
        """
        Set current entry point url

        :return:
        """
        self._current_url = url
        # update preference
        self.preferences.set(
            INDEXERS_CURRENT_ENTRY_POINT_URL_PREFERENCES_KEY,
            self._current_url,
        )



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from hashlib import sha256
from typing import Optional, TypeVar
from uuid import UUID

import base58

from tikka.libs.keypair import ss58_decode

AccountType = TypeVar("AccountType", bound="Account")


class AccountCryptoType(int, Enum):
    ED25519 = 0
    SR25519 = 1
    ECDSA = 2


@dataclass
class AccountBalance:

    total: int
    available: int
    reserved: int


@dataclass
class Account:

    address: str
    name: Optional[str] = None
    crypto_type: AccountCryptoType = AccountCryptoType.ED25519
    balance: Optional[int] = None
    balance_available: Optional[int] = None
    balance_reserved: Optional[int] = None
    path: Optional[str] = None
    root: Optional[str] = None
    file_import: bool = False
    category_id: Optional[UUID] = None
    legacy_v1: bool = False
    total_transfers_count: int = 0
    last_transfer_timestamp: Optional[datetime] = None
    oldest_transfer_timestamp: Optional[datetime] = None

    def __str__(self):
        """
        Return string representation

        :return:
        """
        if self.name:
            return f"{self.address} - {self.name}"
        return f"{self.address}"

    def __eq__(self, other):
        """
        Test equality on address

        :param other: Account instance
        :return:
        """
        if not isinstance(other, self.__class__):
            return False
        return other.address == self.address

    def __hash__(self):
        return hash(self.address)

    def get_v1_address(self, ss58_format: int):
        """
        Return address in V1 format (Base58 of pubkey)

        :return:
        """
        pubkey_hex = ss58_decode(self.address, ss58_format)
        checksum = base58.b58encode(
            sha256(sha256(bytes.fromhex(pubkey_hex)).digest()).digest()
        ).decode("utf8")[:3]
        v1_address = base58.b58encode(bytes.fromhex(pubkey_hex)).decode("utf8")
        return f"{v1_address}:{checksum}"



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from gql import gql

from tikka.adapters.network.indexer.accounts import IndexerAccounts
from tikka.adapters.network.indexer.connection import IndexerConnection
from tikka.adapters.network.indexer.identities import IndexerIdentities
from tikka.adapters.network.indexer.transfers import IndexerTransfers
from tikka.domains.entities.indexer import Indexer
from tikka.interfaces.adapters.network.connection import NetworkConnectionError
from tikka.interfaces.adapters.network.indexer.indexer import (
    NetworkIndexerException,
    NetworkIndexerInterface,
)


class NetworkIndexer(NetworkIndexerInterface):
    """
    NetworkIndexer class
    """

    def __init__(self):
        """
        Init NetworkIndexer instance
        """
        self._connection = IndexerConnection()
        self._transfers = IndexerTransfers(self)
        self._identities = IndexerIdentities(self)
        self._accounts = IndexerAccounts(self)

    @property
    def connection(self) -> IndexerConnection:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkIndexerInterface.connection.__doc__
        )
        return self._connection

    def get(self) -> Indexer:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkIndexerInterface.get.__doc__
        )
        if not self.connection.is_connected() or self.connection.client is None:
            raise NetworkIndexerException(NetworkConnectionError())

        query = gql(
            """
            query {
                blocks (
                    orderBy: HEIGHT_DESC
                    first: 1
                ) {
                    nodes {
                        height
                    }
                }
            }
            """
        )
        try:
            result = self.connection.client.execute(query)
        except Exception as exception:
            raise NetworkIndexerException(exception)

        last_block = result["blocks"]["nodes"][0]["height"]
        return Indexer(self.connection.url, block=last_block)

    def get_genesis_hash(self) -> str:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkIndexerInterface.get_genesis_hash.__doc__
        )
        if not self.connection.is_connected() or self.connection.client is None:
            raise NetworkIndexerException(NetworkConnectionError())

        query = gql(
            """
            query {
                blocks (
                    filter: {
                      height: {
                        equalTo: 0
                      }
                      }
                    
                ) {
                    nodes {
                        hash
                    }
                }
            }
            """
        )
        try:
            result = self.connection.client.execute(query)
        except Exception as exception:
            raise NetworkIndexerException(exception)
        raw_hash = result["blocks"]["nodes"][0]["hash"]
        return f"0{raw_hash[1:]}"

    @property
    def identities(self) -> IndexerIdentities:
        """
        Return IndexerIdentities instance

        :return:
        """
        return self._identities

    @property
    def transfers(self) -> IndexerTransfers:
        """
        Return IndexerTransfers instance

        :return:
        """
        return self._transfers

    @property
    def accounts(self) -> IndexerAccounts:
        """
        Return IndexerAccounts instance

        :return:
        """
        return self._accounts



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/windows/address_add.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AddressAddDialog(object):
    def setupUi(self, AddressAddDialog):
        AddressAddDialog.setObjectName("AddressAddDialog")
        AddressAddDialog.resize(629, 394)
        self.verticalLayout = QtWidgets.QVBoxLayout(AddressAddDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.imageLabel = QtWidgets.QLabel(AddressAddDialog)
        self.imageLabel.setMaximumSize(QtCore.QSize(200, 200))
        self.imageLabel.setText("")
        self.imageLabel.setPixmap(QtGui.QPixmap(":/images/safe_girl"))
        self.imageLabel.setScaledContents(True)
        self.imageLabel.setObjectName("imageLabel")
        self.verticalLayout.addWidget(self.imageLabel, 0, QtCore.Qt.AlignHCenter)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.balanceLabel = QtWidgets.QLabel(AddressAddDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.balanceLabel.setFont(font)
        self.balanceLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.balanceLabel.setObjectName("balanceLabel")
        self.gridLayout.addWidget(self.balanceLabel, 2, 0, 1, 1)
        self.balanceValueLabel = QtWidgets.QLabel(AddressAddDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.balanceValueLabel.sizePolicy().hasHeightForWidth())
        self.balanceValueLabel.setSizePolicy(sizePolicy)
        self.balanceValueLabel.setStyleSheet("color: white;\n"
"border : 3px solid rgb(34, 156, 244);\n"
"border-radius: 10%;\n"
"background: rgb(34, 156, 244);\n"
"\n"
"\n"
"")
        self.balanceValueLabel.setObjectName("balanceValueLabel")
        self.gridLayout.addWidget(self.balanceValueLabel, 2, 1, 1, 1)
        self.addressLineEdit = QtWidgets.QLineEdit(AddressAddDialog)
        self.addressLineEdit.setReadOnly(False)
        self.addressLineEdit.setObjectName("addressLineEdit")
        self.gridLayout.addWidget(self.addressLineEdit, 0, 1, 1, 1)
        self.loaderIconLabel = QtWidgets.QLabel(AddressAddDialog)
        self.loaderIconLabel.setMaximumSize(QtCore.QSize(16, 16))
        self.loaderIconLabel.setText("")
        self.loaderIconLabel.setScaledContents(True)
        self.loaderIconLabel.setObjectName("loaderIconLabel")
        self.gridLayout.addWidget(self.loaderIconLabel, 0, 2, 1, 1)
        self.keyTypeLabel = QtWidgets.QLabel(AddressAddDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.keyTypeLabel.setFont(font)
        self.keyTypeLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.keyTypeLabel.setObjectName("keyTypeLabel")
        self.gridLayout.addWidget(self.keyTypeLabel, 1, 0, 1, 1)
        self.keyTypeValueLabel = QtWidgets.QLabel(AddressAddDialog)
        self.keyTypeValueLabel.setObjectName("keyTypeValueLabel")
        self.gridLayout.addWidget(self.keyTypeValueLabel, 1, 1, 1, 1)
        self.addressLabel = QtWidgets.QLabel(AddressAddDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.addressLabel.setFont(font)
        self.addressLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.addressLabel.setObjectName("addressLabel")
        self.gridLayout.addWidget(self.addressLabel, 0, 0, 1, 1)
        self.nameLabel = QtWidgets.QLabel(AddressAddDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.nameLabel.setFont(font)
        self.nameLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.nameLabel.setObjectName("nameLabel")
        self.gridLayout.addWidget(self.nameLabel, 3, 0, 1, 1)
        self.nameLineEdit = QtWidgets.QLineEdit(AddressAddDialog)
        self.nameLineEdit.setObjectName("nameLineEdit")
        self.gridLayout.addWidget(self.nameLineEdit, 3, 1, 1, 1)
        self.identityIconLabel = QtWidgets.QLabel(AddressAddDialog)
        self.identityIconLabel.setMaximumSize(QtCore.QSize(16, 16))
        self.identityIconLabel.setText("")
        self.identityIconLabel.setObjectName("identityIconLabel")
        self.gridLayout.addWidget(self.identityIconLabel, 2, 2, 1, 1)
        self.identityNameAndIndexValuelabel = QtWidgets.QLabel(AddressAddDialog)
        self.identityNameAndIndexValuelabel.setStyleSheet("color: blue")
        self.identityNameAndIndexValuelabel.setText("")
        self.identityNameAndIndexValuelabel.setObjectName("identityNameAndIndexValuelabel")
        self.gridLayout.addWidget(self.identityNameAndIndexValuelabel, 3, 2, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.errorLabel = QtWidgets.QLabel(AddressAddDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.errorLabel.sizePolicy().hasHeightForWidth())
        self.errorLabel.setSizePolicy(sizePolicy)
        self.errorLabel.setStyleSheet("color: rgb(255, 0, 0)")
        self.errorLabel.setObjectName("errorLabel")
        self.verticalLayout.addWidget(self.errorLabel, 0, QtCore.Qt.AlignHCenter)
        self.buttonBox = QtWidgets.QDialogButtonBox(AddressAddDialog)
        self.buttonBox.setEnabled(True)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox, 0, QtCore.Qt.AlignHCenter)

        self.retranslateUi(AddressAddDialog)
        self.buttonBox.accepted.connect(AddressAddDialog.accept) # type: ignore
        self.buttonBox.rejected.connect(AddressAddDialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(AddressAddDialog)

    def retranslateUi(self, AddressAddDialog):
        #  _translate = QtCore.QCoreApplication.translate
        AddressAddDialog.setWindowTitle(_("Add an address"))
        self.balanceLabel.setText(_("Balance"))
        self.balanceValueLabel.setText(_("?"))
        self.keyTypeLabel.setText(_("Key type"))
        self.keyTypeValueLabel.setText(_("?"))
        self.addressLabel.setText(_("Address"))
        self.nameLabel.setText(_("Add a name"))
import tikka.slots.pyqt.resources.icons.index_rc



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from tikka.domains.entities.account import Account


class AccountsRepositoryInterface(abc.ABC):
    """
    AccountRepositoryInterface class
    """

    COLUMN_ADDRESS = "account_address"
    COLUMN_NAME = "account_name"
    COLUMN_CRYPTO_TYPE = "account_crypto_type"
    COLUMN_BALANCE = "account_balance"
    COLUMN_PATH = "account_path"
    COLUMN_ROOT = "account_root"
    COLUMN_FILE_IMPORT = "account_file_import"
    COLUMN_CATEGORY_ID = "account_category_id"
    COLUMN_LEGACY_V1 = "account_legacy_v1"
    COLUMN_TOTAL_TRANSFERS_COUNT = "account_total_transfers_count"

    SORT_ORDER_ASCENDING = "ASC"
    SORT_ORDER_DESCENDING = "DESC"

    TABLE_VIEW_FILTER_BY_CATEGORY_ID = "table_view_filter_category_id"
    TABLE_VIEW_FILTER_BY_WALLET = "table_view_filter_wallet"

    @dataclass
    class TableViewRow:
        wallet_address: Optional[str]
        identity_index: Optional[int]
        identity_name: Optional[str]
        balance: Optional[int]
        name: Optional[str]
        address: str
        path: Optional[str]
        root: Optional[str]
        crypto_type: int
        legacy_v1: bool
        category_name: Optional[str]

    @abc.abstractmethod
    def add(self, account: Account) -> None:
        """
        Add a new account in repository

        :param account: Account instance
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def list(
        self,
        filters: Optional[Dict[str, Any]] = None,
        sort_column: Optional[str] = None,
        sort_order: str = SORT_ORDER_ASCENDING,
    ) -> List[Account]:
        """
        List accounts from repository with optional filters and sort_column

        :param filters: Dict with {column: value} filters or None
        :param sort_column: Sort column constant like COLUMN_ADDRESS or None
        :param sort_order: Sort order constant SORT_ORDER_ASCENDING or SORT_ORDER_DESCENDING
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def table_view(
        self,
        filters: Optional[Dict[str, Any]] = None,
        sort_column_index: Optional[int] = None,
        sort_order: str = SORT_ORDER_ASCENDING,
    ) -> List[TableViewRow]:
        """
        List accounts from repository with optional filters and sort_column

        :param filters: Dict with {column: value} filters or None
        :param sort_column_index: Sort column index or None
        :param sort_order: Sort order constant SORT_ORDER_ASCENDING or SORT_ORDER_DESCENDING
        :return:
        """
        raise NotImplementedError

    def update(self, account: Account) -> None:
        """
        Update account in repository

        :param account: Account instance
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, address: str) -> None:
        """
        Delete account with address in repository

        :param address: Account address
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete_all(self) -> None:
        """
        Delete all accounts

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def count(self) -> int:
        """
        Return total number of accounts

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def total_balance(self) -> int:
        """
        Return total sum of all account balances

        :return:
        """
        raise NotImplementedError



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import List, Optional

from PyQt5.QtCore import QAbstractTableModel, QModelIndex, Qt, QVariant

from tikka.domains.application import Application
from tikka.domains.entities.node import Node
from tikka.interfaces.adapters.repository.nodes import NodesRepositoryInterface
from tikka.slots.pyqt.entities.constants import (
    NODES_TABLE_SORT_COLUMN_PREFERENCES_KEY,
    NODES_TABLE_SORT_ORDER_PREFERENCES_KEY,
)


class NodesTableModel(QAbstractTableModel):
    """
    NodesTableModel class that drives the population of tabular display
    """

    def __init__(self, application: Application):
        super().__init__()

        self.application = application
        self._ = self.application.translator.gettext

        self.headers = [
            self._("Url"),
            self._("Software"),
            self._("Version"),
            self._("Peer ID"),
            self._("Unsafe API"),
        ]

        self.column_types = [
            NodesRepositoryInterface.COLUMN_URL,
            NodesRepositoryInterface.COLUMN_SOFTWARE,
            NodesRepositoryInterface.COLUMN_SOFTWARE_VERSION,
            NodesRepositoryInterface.COLUMN_PEER_ID,
            NodesRepositoryInterface.COLUMN_UNSAFE_API_EXPOSED,
        ]
        self.sort_order_types = [
            NodesRepositoryInterface.SORT_ORDER_ASCENDING,
            NodesRepositoryInterface.SORT_ORDER_DESCENDING,
        ]

        self.nodes: List[Node] = []
        self.init_data()

    def init_data(self):
        """
        Fill data from repository

        :return:
        """
        pref_sort_column = self.application.repository.preferences.get(
            NODES_TABLE_SORT_COLUMN_PREFERENCES_KEY
        )
        if pref_sort_column is None:
            sort_column = NodesRepositoryInterface.COLUMN_URL
        else:
            sort_column = self.column_types[int(pref_sort_column)]

        pref_sort_order = self.application.repository.preferences.get(
            NODES_TABLE_SORT_ORDER_PREFERENCES_KEY
        )
        if pref_sort_order is None:
            sort_order = NodesRepositoryInterface.SORT_ORDER_ASCENDING
        else:
            sort_order = self.sort_order_types[int(pref_sort_order)]
        self.beginResetModel()
        self.nodes = self.application.nodes.repository.list(
            sort_column=sort_column,
            sort_order=sort_order,
        )
        self.endResetModel()

    def sort(self, column: int, order: Optional[Qt.SortOrder] = None):
        """
        Triggered by Qt Signal Sort by column

        :param column: Index of sort column
        :param order: Qt.SortOrder flag
        :return:
        """
        if column > -1:
            self.application.repository.preferences.set(
                NODES_TABLE_SORT_COLUMN_PREFERENCES_KEY, str(column)
            )
            self.application.repository.preferences.set(
                NODES_TABLE_SORT_ORDER_PREFERENCES_KEY, str(order)
            )
        self.init_data()

    def rowCount(self, _: QModelIndex = QModelIndex()) -> int:
        """
        Return row count

        :param _: QModelIndex instance
        :return:
        """
        count = self.application.nodes.repository.count()
        if count == 0:
            return 0
        if count <= len(self.nodes):
            return count

        return len(self.nodes)

    def columnCount(self, _: QModelIndex = QModelIndex()) -> int:
        """
        Return column count (length of headers list)

        :param _: QModelIndex instance
        :return:
        """
        return len(self.headers)

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole) -> QVariant:
        """
        Return data of cell for column index.column

        :param index: QModelIndex instance
        :param role: Item data role
        :return:
        """
        col = index.column()
        row = index.row()
        node = self.nodes[row]
        data = QVariant()
        if role == Qt.DisplayRole:
            if col == 0:
                data = QVariant(node.url)
            if col == 1:
                data = QVariant(node.software)
            if col == 2:
                data = QVariant(node.software_version)
            if col == 3:
                data = QVariant(node.peer_id)
            if col == 4:
                data = QVariant(
                    self._("Exposed") if node.unsafe_api_exposed is True else ""
                )
        return data

    def headerData(
        self, section: int, orientation: int, role: int = Qt.DisplayRole
    ) -> QVariant:
        """
        Return

        :param section: Headers column index
        :param orientation: Headers orientation
        :param role: Item role
        :return:
        """
        if role != Qt.DisplayRole:
            return QVariant()

        if orientation == Qt.Horizontal:
            return QVariant(self.headers[section])

        # return row number as vertical header
        return QVariant(int(section + 1))



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Optional

from PyQt5.QtCore import QMutex, Qt, QTimer
from PyQt5.QtGui import QFont, QMovie, QPixmap
from PyQt5.QtWidgets import QApplication, QDialog, QLineEdit, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.account import Account, AccountBalance, AccountCryptoType
from tikka.domains.entities.constants import (
    AMOUNT_UNIT_KEY,
    DATA_PATH,
    WALLETS_PASSWORD_LENGTH,
)
from tikka.domains.entities.identity import Identity, IdentityStatus
from tikka.domains.entities.smith import Smith, SmithStatus
from tikka.libs.keypair import Keypair
from tikka.libs.secret import generate_alphabetic
from tikka.libs.signing_key_v1 import SigningKey
from tikka.slots.pyqt.entities.constants import (
    ADDRESS_MONOSPACE_FONT_NAME,
    DEBOUNCE_TIME,
    ICON_IDENTITY_MEMBER,
    ICON_IDENTITY_MEMBER_NOT_VALIDATED,
    ICON_IDENTITY_NOT_MEMBER,
    ICON_LOADER,
    ICON_SMITH,
    SELECTED_UNIT_PREFERENCES_KEY,
)
from tikka.slots.pyqt.entities.worker import AsyncQWorker
from tikka.slots.pyqt.resources.gui.windows.v1_account_import_rc import (
    Ui_V1AccountImportDialog,
)


class V1AccountImportWindow(QDialog, Ui_V1AccountImportDialog):
    """
    V1AccountImportWindow class
    """

    def __init__(
        self, application: Application, mutex: QMutex, parent: Optional[QWidget] = None
    ):
        """
        Init import V1 account window

        :param application: Application instance
        :param mutex: QMutex instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext
        self.mutex = mutex

        self.is_legacy_v1: Optional[bool] = None
        self.balance: Optional[AccountBalance] = None
        self.identity: Optional[Identity] = None
        self.smith: Optional[Smith] = None

        self.display_identity_icon = {
            IdentityStatus.UNCONFIRMED.value: QPixmap(
                ICON_IDENTITY_MEMBER_NOT_VALIDATED
            ),
            IdentityStatus.UNVALIDATED.value: QPixmap(
                ICON_IDENTITY_MEMBER_NOT_VALIDATED
            ),
            IdentityStatus.MEMBER.value: QPixmap(ICON_IDENTITY_MEMBER),
            IdentityStatus.NOT_MEMBER.value: QPixmap(ICON_IDENTITY_NOT_MEMBER),
            IdentityStatus.REVOKED.value: QPixmap(ICON_IDENTITY_NOT_MEMBER),
        }

        # calculate balance display by unit preference
        unit_preference = self.application.repository.preferences.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            self.amount = self.application.amounts.get_amount(unit_preference)
        else:
            self.amount = self.application.amounts.get_amount(AMOUNT_UNIT_KEY)

        # set monospace font to address fields
        monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        monospace_font.setStyleHint(QFont.Monospace)
        self.addressValueLabel.setFont(monospace_font)
        self.v1AddressValueLabel.setFont(monospace_font)

        # animated loading icon
        self.loader_movie = QMovie(ICON_LOADER)
        self.loader_movie.start()
        self.loaderIconLabel.setMovie(self.loader_movie)
        loader_icon_size_policy = self.loaderIconLabel.sizePolicy()
        loader_icon_size_policy.setRetainSizeWhenHidden(True)
        self.loaderIconLabel.setSizePolicy(loader_icon_size_policy)
        self.loaderIconLabel.hide()

        # buttons
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

        # events
        self.secretIDLineEdit.textChanged.connect(self._on_secret_id_line_edit_changed)
        self.passwordIDLineEdit.textChanged.connect(
            self._on_password_id_line_edit_changed
        )

        self.showButton.clicked.connect(self.on_show_button_clicked)
        self.passwordChangeButton.clicked.connect(self._generate_wallet_password)
        self.buttonBox.accepted.connect(self.on_accepted_button)
        self.buttonBox.rejected.connect(self.close)

        # debounce timer on self._generate_address()
        self.debounce_timer = QTimer()
        self.debounce_timer.timeout.connect(self._generate_address)
        # Create a QWorker object
        self.network_fetch_account_data_async_qworker = AsyncQWorker(
            self.network_fetch_account_data, self.mutex
        )
        self.network_fetch_account_data_async_qworker.finished.connect(
            self._on_finished_network_fetch_account_data
        )

        # fill form
        self._generate_wallet_password()

    def _on_secret_id_line_edit_changed(self):
        """
        Triggered when text is changed in the secret ID field

        :return:
        """
        if self.debounce_timer.isActive():
            self.debounce_timer.stop()
        self.debounce_timer.start(DEBOUNCE_TIME)

    def _on_password_id_line_edit_changed(self):
        """
        Triggered when text is changed in the password ID field

        :return:
        """
        if self.debounce_timer.isActive():
            self.debounce_timer.stop()
        self.debounce_timer.start(DEBOUNCE_TIME)

    def _generate_address(self):
        """
        Generate address from ID

        :return:
        """
        # stop debounce_timer to avoid infinite loop
        if self.debounce_timer.isActive():
            self.debounce_timer.stop()

        self.v1AddressValueLabel.setText("")
        self.addressValueLabel.setText("")
        self.errorLabel.setText("")
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

        secret_id = self.secretIDLineEdit.text().strip()
        password_id = self.passwordIDLineEdit.text().strip()
        if secret_id == "" or password_id == "":
            return

        signing_key = SigningKey.from_credentials(secret_id, password_id)
        try:
            address = Keypair.create_from_seed(
                seed_hex=signing_key.seed.hex(),
                ss58_format=self.application.currencies.get_current().ss58_format,
                crypto_type=AccountCryptoType.ED25519,
            ).ss58_address
        except Exception as exception:
            logging.exception(exception)
            self.errorLabel.setText(self._("Error generating account wallet!"))
            return

        self.addressValueLabel.setText(address)
        self.v1AddressValueLabel.setText(
            Account(address).get_v1_address(
                self.application.currencies.get_current().ss58_format
            )
        )
        if self.application.connections.node.is_connected():
            self.loaderIconLabel.show()
            self.network_fetch_account_data_async_qworker.start()

    def on_show_button_clicked(self):
        """
        Triggered when user click on show button

        :return:
        """
        if self.secretIDLineEdit.echoMode() == QLineEdit.Password:
            self.secretIDLineEdit.setEchoMode(QLineEdit.Normal)
            self.passwordIDLineEdit.setEchoMode(QLineEdit.Normal)
            self.showButton.setText(self._("Hide"))
        else:
            self.secretIDLineEdit.setEchoMode(QLineEdit.Password)
            self.passwordIDLineEdit.setEchoMode(QLineEdit.Password)
            self.showButton.setText(self._("Show"))

    def _generate_wallet_password(self):
        """
        Generate new password for wallet encryption in UI

        :return:
        """
        self.passwordLineEdit.setText(generate_alphabetic(WALLETS_PASSWORD_LENGTH))

    def network_fetch_account_data(self):
        """
        Fetch account data and
        check if account is really a known legacy v1 account

        :return:
        """
        address = self.addressValueLabel.text().strip()
        self.balance = None
        self.identity = None
        self.smith = None
        try:
            self.is_legacy_v1 = self.application.accounts.network_is_legacy_v1(address)
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)
        else:
            try:
                self.balance = self.application.accounts.network_get_balance(address)
            except Exception as exception:
                self.errorLabel.setText(self._(str(exception)))
                logging.exception(exception)
            else:
                try:
                    self.identity = self.application.identities.network_get_identity(
                        address
                    )
                except Exception as exception:
                    self.errorLabel.setText(self._(str(exception)))
                    logging.exception(exception)
                else:
                    if self.identity is not None:
                        try:
                            self.smith = self.application.smiths.network_get_smith(
                                self.identity.index
                            )
                        except Exception as exception:
                            self.errorLabel.setText(self._(str(exception)))
                            logging.exception(exception)

    def _on_finished_network_fetch_account_data(self):
        """
        Triggered when async request network_fetch_account_data_async_qworker is finished

        :return:
        """
        self.loaderIconLabel.hide()

        if self.is_legacy_v1:
            address = self.addressValueLabel.text().strip()
            if self.application.accounts.get_by_address(address) is not None:
                self.errorLabel.setText(self._("Account already exists!"))
                self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            else:
                self.errorLabel.setText("")
                self.buttonBox.button(self.buttonBox.Ok).setEnabled(True)
            if self.balance is not None:
                self.balanceValueLabel.setText(
                    self.locale().toCurrencyString(
                        self.amount.value(self.balance.total), self.amount.symbol()
                    )
                )
            else:
                self.balanceValueLabel.setText("?")
            if self.identity:
                self.identityIconLabel.setPixmap(
                    self.display_identity_icon[self.identity.status.value].scaled(
                        16, 18, aspectRatioMode=Qt.KeepAspectRatio
                    )
                )
                identity_name = self.identity.name or ""
                self.identityNameAndIndexValuelabel.setText(
                    f"{identity_name}#{self.identity.index}"
                )
                if self.smith is not None and self.smith.status == SmithStatus.SMITH:
                    self.identityIconLabel.setPixmap(
                        QPixmap(ICON_SMITH).scaled(
                            16, 18, aspectRatioMode=Qt.KeepAspectRatio
                        )
                    )
            else:
                self.identityNameAndIndexValuelabel.setText("")
                self.identityIconLabel.setPixmap(QPixmap())
        else:
            self.errorLabel.setText(self._("V1 account unknown!"))
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

    def on_accepted_button(self):
        """
        Triggered when user click on ok button

        :return:
        """
        secret_id = self.secretIDLineEdit.text().strip()
        password_id = self.passwordIDLineEdit.text().strip()
        password = self.passwordLineEdit.text()
        name = self.nameLineEdit.text().strip()

        account = self.application.accounts.create_new_root_account_v1_from_credentials(
            secret_id, password_id, name, password
        )
        if account:
            if self.balance:
                account.balance = self.balance.total
                account.balance_available = self.balance.available
                account.balance_reserved = self.balance.reserved
                self.application.accounts.update(account)

            if self.identity:
                if self.application.identities.get(self.identity.index) is None:
                    self.application.identities.add(self.identity)
                else:
                    self.application.identities.update(self.identity)
                if self.smith:
                    if self.application.smiths.get(self.identity.index) is None:
                        self.application.smiths.add(self.smith)
                    else:
                        self.application.smiths.update(self.smith)


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    V1AccountImportWindow(application_, QMutex()).exec_()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/widgets/account.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AccountWidget(object):
    def setupUi(self, AccountWidget):
        AccountWidget.setObjectName("AccountWidget")
        AccountWidget.resize(841, 401)
        AccountWidget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(AccountWidget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.QRCodeAddressLabel = QtWidgets.QLabel(AccountWidget)
        self.QRCodeAddressLabel.setMaximumSize(QtCore.QSize(100, 100))
        self.QRCodeAddressLabel.setBaseSize(QtCore.QSize(200, 0))
        self.QRCodeAddressLabel.setText("")
        self.QRCodeAddressLabel.setPixmap(QtGui.QPixmap(":/images/safe"))
        self.QRCodeAddressLabel.setScaledContents(True)
        self.QRCodeAddressLabel.setObjectName("QRCodeAddressLabel")
        self.horizontalLayout.addWidget(self.QRCodeAddressLabel)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.balanceValueLabel = QtWidgets.QLabel(AccountWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.balanceValueLabel.sizePolicy().hasHeightForWidth())
        self.balanceValueLabel.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setPointSize(40)
        self.balanceValueLabel.setFont(font)
        self.balanceValueLabel.setStyleSheet("color: white;\n"
"border : 3px solid rgb(34, 156, 244);\n"
"border-radius: 25%;\n"
"background: rgb(34, 156, 244);\n"
"\n"
"\n"
"")
        self.balanceValueLabel.setObjectName("balanceValueLabel")
        self.verticalLayout_3.addWidget(self.balanceValueLabel)
        self.balanceReservedValueLabel = QtWidgets.QLabel(AccountWidget)
        self.balanceReservedValueLabel.setStyleSheet("color: orange")
        self.balanceReservedValueLabel.setText("")
        self.balanceReservedValueLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.balanceReservedValueLabel.setObjectName("balanceReservedValueLabel")
        self.verticalLayout_3.addWidget(self.balanceReservedValueLabel)
        self.horizontalLayout.addLayout(self.verticalLayout_3)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.refreshButton = QtWidgets.QPushButton(AccountWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.refreshButton.sizePolicy().hasHeightForWidth())
        self.refreshButton.setSizePolicy(sizePolicy)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/refresh_icon"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.refreshButton.setIcon(icon)
        self.refreshButton.setObjectName("refreshButton")
        self.verticalLayout.addWidget(self.refreshButton, 0, QtCore.Qt.AlignLeft)
        self.transferToButton = QtWidgets.QPushButton(AccountWidget)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/icons/money-recive-svgrepo-com.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.transferToButton.setIcon(icon1)
        self.transferToButton.setIconSize(QtCore.QSize(32, 32))
        self.transferToButton.setObjectName("transferToButton")
        self.verticalLayout.addWidget(self.transferToButton)
        self.horizontalLayout.addLayout(self.verticalLayout)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.formLayout_2 = QtWidgets.QFormLayout()
        self.formLayout_2.setContentsMargins(6, -1, 0, -1)
        self.formLayout_2.setObjectName("formLayout_2")
        self.lockStatusIcon = QtWidgets.QLabel(AccountWidget)
        self.lockStatusIcon.setMaximumSize(QtCore.QSize(16, 16))
        self.lockStatusIcon.setText("")
        self.lockStatusIcon.setPixmap(QtGui.QPixmap(":/images/locked_cartoon_400x400.png"))
        self.lockStatusIcon.setScaledContents(True)
        self.lockStatusIcon.setObjectName("lockStatusIcon")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.lockStatusIcon)
        self.addressValueLabel = QtWidgets.QLabel(AccountWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.addressValueLabel.sizePolicy().hasHeightForWidth())
        self.addressValueLabel.setSizePolicy(sizePolicy)
        self.addressValueLabel.setText("")
        self.addressValueLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.addressValueLabel.setObjectName("addressValueLabel")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.addressValueLabel)
        self.derivationValueLabel = QtWidgets.QLabel(AccountWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.derivationValueLabel.sizePolicy().hasHeightForWidth())
        self.derivationValueLabel.setSizePolicy(sizePolicy)
        self.derivationValueLabel.setText("")
        self.derivationValueLabel.setObjectName("derivationValueLabel")
        self.formLayout_2.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.derivationValueLabel)
        self.derivationIconLabel = QtWidgets.QLabel(AccountWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.derivationIconLabel.sizePolicy().hasHeightForWidth())
        self.derivationIconLabel.setSizePolicy(sizePolicy)
        self.derivationIconLabel.setMaximumSize(QtCore.QSize(16, 20))
        self.derivationIconLabel.setText("")
        self.derivationIconLabel.setPixmap(QtGui.QPixmap(":/icons/wot_icon"))
        self.derivationIconLabel.setScaledContents(True)
        self.derivationIconLabel.setObjectName("derivationIconLabel")
        self.formLayout_2.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.derivationIconLabel)
        self.v1AddressValueLabel = QtWidgets.QLabel(AccountWidget)
        font = QtGui.QFont()
        font.setItalic(True)
        self.v1AddressValueLabel.setFont(font)
        self.v1AddressValueLabel.setText("")
        self.v1AddressValueLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.v1AddressValueLabel.setObjectName("v1AddressValueLabel")
        self.formLayout_2.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.v1AddressValueLabel)
        self.horizontalLayout_3.addLayout(self.formLayout_2)
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.identityStatusIconLabel = QtWidgets.QLabel(AccountWidget)
        self.identityStatusIconLabel.setMaximumSize(QtCore.QSize(16, 20))
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.identityStatusIconLabel.setFont(font)
        self.identityStatusIconLabel.setText("")
        self.identityStatusIconLabel.setPixmap(QtGui.QPixmap(":/icons/member_icon"))
        self.identityStatusIconLabel.setScaledContents(True)
        self.identityStatusIconLabel.setObjectName("identityStatusIconLabel")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.identityStatusIconLabel)
        self.identityValueLabel = QtWidgets.QLabel(AccountWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.identityValueLabel.sizePolicy().hasHeightForWidth())
        self.identityValueLabel.setSizePolicy(sizePolicy)
        self.identityValueLabel.setObjectName("identityValueLabel")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.identityValueLabel)
        self.smithLabel = QtWidgets.QLabel(AccountWidget)
        self.smithLabel.setMaximumSize(QtCore.QSize(16, 20))
        self.smithLabel.setText("")
        self.smithLabel.setPixmap(QtGui.QPixmap(":/icons/smith"))
        self.smithLabel.setScaledContents(True)
        self.smithLabel.setObjectName("smithLabel")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.smithLabel)
        self.smithValuelabel = QtWidgets.QLabel(AccountWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.smithValuelabel.sizePolicy().hasHeightForWidth())
        self.smithValuelabel.setSizePolicy(sizePolicy)
        self.smithValuelabel.setObjectName("smithValuelabel")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.smithValuelabel)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setSpacing(0)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.identityNameValueLabel = QtWidgets.QLabel(AccountWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.identityNameValueLabel.sizePolicy().hasHeightForWidth())
        self.identityNameValueLabel.setSizePolicy(sizePolicy)
        self.identityNameValueLabel.setStyleSheet("color: blue")
        self.identityNameValueLabel.setObjectName("identityNameValueLabel")
        self.horizontalLayout_4.addWidget(self.identityNameValueLabel)
        self.accountNameValueLabel = QtWidgets.QLabel(AccountWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.accountNameValueLabel.sizePolicy().hasHeightForWidth())
        self.accountNameValueLabel.setSizePolicy(sizePolicy)
        self.accountNameValueLabel.setObjectName("accountNameValueLabel")
        self.horizontalLayout_4.addWidget(self.accountNameValueLabel)
        self.formLayout.setLayout(0, QtWidgets.QFormLayout.FieldRole, self.horizontalLayout_4)
        self.horizontalLayout_3.addLayout(self.formLayout)
        self.verticalLayout_2.addLayout(self.horizontalLayout_3)
        self.errorLabel = QtWidgets.QLabel(AccountWidget)
        self.errorLabel.setStyleSheet("color: red")
        self.errorLabel.setText("")
        self.errorLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.errorLabel.setObjectName("errorLabel")
        self.verticalLayout_2.addWidget(self.errorLabel)
        self.groupBox = QtWidgets.QGroupBox(AccountWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox.sizePolicy().hasHeightForWidth())
        self.groupBox.setSizePolicy(sizePolicy)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.transfersCountValueLabel = QtWidgets.QLabel(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.transfersCountValueLabel.sizePolicy().hasHeightForWidth())
        self.transfersCountValueLabel.setSizePolicy(sizePolicy)
        self.transfersCountValueLabel.setText("")
        self.transfersCountValueLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.transfersCountValueLabel.setObjectName("transfersCountValueLabel")
        self.horizontalLayout_2.addWidget(self.transfersCountValueLabel)
        self.verticalLayout_4.addLayout(self.horizontalLayout_2)
        self.transfersListView = QtWidgets.QListView(self.groupBox)
        self.transfersListView.setMouseTracking(True)
        self.transfersListView.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.transfersListView.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.transfersListView.setAlternatingRowColors(True)
        self.transfersListView.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
        self.transfersListView.setObjectName("transfersListView")
        self.verticalLayout_4.addWidget(self.transfersListView)
        self.verticalLayout_2.addWidget(self.groupBox)

        self.retranslateUi(AccountWidget)
        QtCore.QMetaObject.connectSlotsByName(AccountWidget)

    def retranslateUi(self, AccountWidget):
        #  _translate = QtCore.QCoreApplication.translate
        AccountWidget.setWindowTitle(_("Account"))
        self.balanceValueLabel.setText(_("?"))
        self.refreshButton.setText(_("Refresh"))
        self.transferToButton.setText(_("Transfer to"))
        self.identityValueLabel.setText(_("None"))
        self.smithValuelabel.setText(_("None"))
        self.identityNameValueLabel.setText(_("identity name"))
        self.accountNameValueLabel.setText(_("account name"))
        self.groupBox.setTitle(_("Transfers"))
import tikka.slots.pyqt.resources.icons.index_rc



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/widgets/technical_committee.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_technicalCommitteeWidget(object):
    def setupUi(self, technicalCommitteeWidget):
        technicalCommitteeWidget.setObjectName("technicalCommitteeWidget")
        technicalCommitteeWidget.resize(811, 653)
        self.verticalLayout = QtWidgets.QVBoxLayout(technicalCommitteeWidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.refreshButton = QtWidgets.QPushButton(technicalCommitteeWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.refreshButton.sizePolicy().hasHeightForWidth())
        self.refreshButton.setSizePolicy(sizePolicy)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/loader"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.refreshButton.setIcon(icon)
        self.refreshButton.setObjectName("refreshButton")
        self.verticalLayout.addWidget(self.refreshButton, 0, QtCore.Qt.AlignRight)
        self.membersGroupBox = QtWidgets.QGroupBox(technicalCommitteeWidget)
        self.membersGroupBox.setObjectName("membersGroupBox")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.membersGroupBox)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.membersListWidget = QtWidgets.QListWidget(self.membersGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.membersListWidget.sizePolicy().hasHeightForWidth())
        self.membersListWidget.setSizePolicy(sizePolicy)
        self.membersListWidget.setObjectName("membersListWidget")
        self.verticalLayout_2.addWidget(self.membersListWidget)
        self.verticalLayout.addWidget(self.membersGroupBox)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.accountLabel = QtWidgets.QLabel(technicalCommitteeWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.accountLabel.setFont(font)
        self.accountLabel.setObjectName("accountLabel")
        self.horizontalLayout.addWidget(self.accountLabel)
        self.accountComboBox = QtWidgets.QComboBox(technicalCommitteeWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.accountComboBox.sizePolicy().hasHeightForWidth())
        self.accountComboBox.setSizePolicy(sizePolicy)
        self.accountComboBox.setObjectName("accountComboBox")
        self.horizontalLayout.addWidget(self.accountComboBox)
        self.loaderIconLabel = QtWidgets.QLabel(technicalCommitteeWidget)
        self.loaderIconLabel.setMaximumSize(QtCore.QSize(16, 16))
        self.loaderIconLabel.setPixmap(QtGui.QPixmap(":/icons/loader"))
        self.loaderIconLabel.setScaledContents(True)
        self.loaderIconLabel.setObjectName("loaderIconLabel")
        self.horizontalLayout.addWidget(self.loaderIconLabel)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.proposalsGroupBox = QtWidgets.QGroupBox(technicalCommitteeWidget)
        self.proposalsGroupBox.setObjectName("proposalsGroupBox")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.proposalsGroupBox)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.proposalsListWidget = QtWidgets.QListWidget(self.proposalsGroupBox)
        self.proposalsListWidget.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
        self.proposalsListWidget.setObjectName("proposalsListWidget")
        self.verticalLayout_3.addWidget(self.proposalsListWidget)
        self.verticalLayout.addWidget(self.proposalsGroupBox)
        self.errorLabel = QtWidgets.QLabel(technicalCommitteeWidget)
        self.errorLabel.setStyleSheet("color: red")
        self.errorLabel.setText("")
        self.errorLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.errorLabel.setObjectName("errorLabel")
        self.verticalLayout.addWidget(self.errorLabel)

        self.retranslateUi(technicalCommitteeWidget)
        QtCore.QMetaObject.connectSlotsByName(technicalCommitteeWidget)

    def retranslateUi(self, technicalCommitteeWidget):
        #  _translate = QtCore.QCoreApplication.translate
        technicalCommitteeWidget.setWindowTitle(_("Form"))
        self.refreshButton.setText(_("Refresh"))
        self.membersGroupBox.setTitle(_("Members"))
        self.accountLabel.setText(_("Voting account"))
        self.proposalsGroupBox.setTitle(_("Proposals"))
import tikka.slots.pyqt.resources.icons.index_rc



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from datetime import datetime
from typing import List, Optional

from dateutil import parser
from gql import gql

from tikka.domains.entities.transfer import Transfer
from tikka.interfaces.adapters.network.connection import NetworkConnectionError
from tikka.interfaces.adapters.network.indexer.transfers import (
    IndexerTransfersException,
    IndexerTransfersInterface,
)


class IndexerTransfers(IndexerTransfersInterface):
    """
    IndexerTransfers class
    """

    def list(
        self,
        address: str,
        limit: int,
        offset: int = 0,
        sort_column: str = IndexerTransfersInterface.COLUMN_TIMESTAMP,
        sort_order: str = IndexerTransfersInterface.SORT_ORDER_DESCENDING,
        from_datetime: Optional[datetime] = None,
        to_datetime: Optional[datetime] = None,
    ) -> List[Transfer]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IndexerTransfersInterface.list.__doc__
        )

        if (
            not self.indexer.connection.is_connected()
            or self.indexer.connection.client is None
        ):
            raise IndexerTransfersException(NetworkConnectionError())

        # Construire la condition de filtrage pour la date
        date_filter = []
        if from_datetime:
            date_filter.append(
                f'{{ timestamp: {{ greaterThanOrEqualTo: "{from_datetime.isoformat()}" }} }}'
            )
        if to_datetime:
            date_filter.append(
                f'{{ timestamp: {{ lessThanOrEqualTo: "{to_datetime.isoformat()}" }} }}'
            )

        date_filter_str = (
            ", and: [" + ", ".join(date_filter) + "]" if date_filter else ""
        )

        # Construire la requête GraphQL
        # Construire la condition de filtrage pour la date
        date_filter = []
        if from_datetime:
            date_filter.append(
                f'{{ timestamp: {{ greaterThanOrEqualTo: "{from_datetime.isoformat()}" }} }}'
            )
        if to_datetime:
            date_filter.append(
                f'{{ timestamp: {{ lessThanOrEqualTo: "{to_datetime.isoformat()}" }} }}'
            )

        date_filter_str = "and: [" + ", ".join(date_filter) + "]" if date_filter else ""

        # Construire la requête GraphQL
        query = f"""
            query {{
              transfers(
                first: {limit}
                offset: {offset}
                orderBy: {sort_column.upper()}_{sort_order.upper()}
                filter: {{
                    or: [
                        {{ fromId: {{ equalTo: "{address}" }} }},
                        {{ toId: {{ equalTo: "{address}" }} }}
                    ]
                    {date_filter_str}
                }}
              ) {{
                nodes {{
                    id
                    fromId
                    from {{
                      identity {{
                        index
                        name
                      }}
                    }}
                    toId
                    to {{
                      identity {{
                        index
                        name
                      }}
                    }}
                    amount
                    timestamp
                    comment {{
                      type
                      remark
                      remarkBytes
                    }}
                }}
              }}
            }}
        """

        try:
            result = self.indexer.connection.client.execute(gql(query))
        except Exception as exception:
            raise IndexerTransfersException(exception)

        transfers = []
        for transfer in result["transfers"]["nodes"]:
            issuer_identity_index = (
                transfer["from"]["identity"]["index"]
                if transfer["from"]["identity"]
                else None
            )
            issuer_identity_name = (
                transfer["from"]["identity"]["name"]
                if transfer["from"]["identity"]
                else None
            )
            receiver_identity_index = (
                transfer["to"]["identity"]["index"]
                if transfer["to"]["identity"]
                else None
            )
            receiver_identity_name = (
                transfer["to"]["identity"]["name"]
                if transfer["to"]["identity"]
                else None
            )

            comment = None
            comment_type = None
            if transfer["comment"]:
                comment = (
                    transfer["comment"]["remarkBytes"]
                    if transfer["comment"]["type"] == "RAW"
                    else transfer["comment"]["remark"]
                )
                comment_type = transfer["comment"]["type"]

            transfers.append(
                Transfer(
                    id=transfer["id"],
                    issuer_address=transfer["fromId"],
                    issuer_identity_index=issuer_identity_index,
                    issuer_identity_name=issuer_identity_name,
                    receiver_address=transfer["toId"],
                    receiver_identity_index=receiver_identity_index,
                    receiver_identity_name=receiver_identity_name,
                    amount=transfer["amount"],
                    timestamp=parser.parse(transfer["timestamp"]),
                    comment=comment,
                    comment_type=comment_type,
                )
            )

        return transfers

    def count(self, address) -> int:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IndexerTransfersInterface.count.__doc__
        )
        if (
            not self.indexer.connection.is_connected()
            or self.indexer.connection.client is None
        ):
            raise IndexerTransfersException(NetworkConnectionError())

        query = f"""
        query {{
          transfers(
            filter: {{
              or: [
                {{ fromId: {{ equalTo: "{address}" }} }}
                {{ toId: {{ equalTo: "{address}" }} }}
              ]
            }}
          ) {{
            totalCount
          }}
        }}
        """
        try:
            result = self.indexer.connection.client.execute(gql(query))
        except Exception as exception:
            raise IndexerTransfersException(exception)

        # {
        #     "data": {
        #         "transfers": {
        #             "totalCount": 916
        #         }
        #     }
        # }
        return result["transfers"]["totalCount"]



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from functools import reduce
from typing import Any, Dict, List, Optional

from dateutil import parser
from sql import Column, Delete, Flavor, Table
from sql.operators import And, Equal

from tikka.domains.entities.transfer import Transfer
from tikka.interfaces.adapters.repository.db_repository import DBRepositoryInterface
from tikka.interfaces.adapters.repository.transfers import TransfersRepositoryInterface

TABLE_NAME = "transfers"
ACCOUNTS_TRANSFERS_TABLE_NAME = "accounts_transfers"

# create sql table wrapper
sql_transfers_table = Table(TABLE_NAME)


class DBTransfersRepository(TransfersRepositoryInterface, DBRepositoryInterface):
    """
    DBTransfersRepository class
    """

    def list(
        self,
        address: str,
        filters: Optional[Dict[str, Any]] = None,
        sort_column: str = TransfersRepositoryInterface.COLUMN_TIMESTAMP,
        sort_order: str = TransfersRepositoryInterface.SORT_ORDER_DESCENDING,
        limit: int = 10,
        offset: int = 0,
    ) -> List[Transfer]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            TransfersRepositoryInterface.list.__doc__
        )

        sql_columns = {
            TransfersRepositoryInterface.COLUMN_ID: sql_transfers_table.id,
            TransfersRepositoryInterface.COLUMN_ISSUER_ADDRESS: sql_transfers_table.issuer_address,
            TransfersRepositoryInterface.COLUMN_ISSUER_IDENTITY_INDEX: sql_transfers_table.issuer_identity_index,
            TransfersRepositoryInterface.COLUMN_ISSUER_IDENTITY_NAME: sql_transfers_table.issuer_identity_name,
            TransfersRepositoryInterface.COLUMN_RECEIVER_ADDRESS: sql_transfers_table.receiver_address,
            TransfersRepositoryInterface.COLUMN_RECEIVER_IDENTITY_INDEX: sql_transfers_table.receiver_identity_index,
            TransfersRepositoryInterface.COLUMN_RECEIVER_IDENTITY_NAME: sql_transfers_table.receiver_identity_name,
            TransfersRepositoryInterface.COLUMN_AMOUNT: sql_transfers_table.amount,
            TransfersRepositoryInterface.COLUMN_TIMESTAMP: sql_transfers_table.timestamp,
            TransfersRepositoryInterface.COLUMN_COMMENT: sql_transfers_table.comment,
            TransfersRepositoryInterface.COLUMN_COMMENT_TYPE: sql_transfers_table.comment_type,
        }

        sql_accounts_transfers_join = sql_transfers_table.join(
            Table(ACCOUNTS_TRANSFERS_TABLE_NAME)
        )
        sql_accounts_transfers_join.condition = Equal(
            sql_accounts_transfers_join.right.account_id, address
        ) & Equal(sql_accounts_transfers_join.right.transfer_id, sql_transfers_table.id)

        # if sort column...
        if sort_column is not None:
            # set sort column
            sql_sort_colum: Column = sql_columns[sort_column]
            # create select query wrapper with order by
            sql_select = sql_accounts_transfers_join.select(
                *sql_columns.values(),
                order_by=sql_sort_colum.asc
                if sort_order == TransfersRepositoryInterface.SORT_ORDER_ASCENDING
                else sql_sort_colum.desc,
                limit=limit,
                offset=offset,
            )
        else:
            #  create select query wrapper without order by
            sql_select = sql_accounts_transfers_join.select(
                *sql_columns.values(), limit=limit, offset=offset
            )

        # create where conditions
        conditions = []
        if filters is not None:
            for key, value in filters.items():
                conditions.append(sql_columns[key] == value)

        # conditions are added with and operator
        def and_(a, b):
            return And((a, b))

        if len(conditions) > 0:
            sql_select.where = reduce(and_, conditions)

        # config sql with ? as param style
        Flavor.set(Flavor(paramstyle="qmark"))

        sql, args = tuple(sql_select)
        result_set = self.client.select(sql, args)

        list_ = []
        for row in result_set:
            list_.append(get_transfer_from_row(row))

        return list_

    def add(self, address: str, transfer: Transfer) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            TransfersRepositoryInterface.add.__doc__
        )

        # insert only non hidden fields
        self.client.insert(
            TABLE_NAME,
            **get_fields_from_transfer(transfer),
        )
        # insert in cross table
        self.client.insert(
            ACCOUNTS_TRANSFERS_TABLE_NAME, account_id=address, transfer_id=transfer.id
        )

    def set_history(self, address: str, transfers: List[Transfer]) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            TransfersRepositoryInterface.set_history.__doc__
        )
        # delete current account history in cross table
        self.client.delete(ACCOUNTS_TRANSFERS_TABLE_NAME, account_id=address)

        # delete orphan transfers not in cross table
        sql_accounts_transfers_table = Table(ACCOUNTS_TRANSFERS_TABLE_NAME)
        sql_delete = Delete(
            sql_transfers_table,
            where=~sql_transfers_table.id.in_(
                sql_accounts_transfers_table.select(
                    sql_accounts_transfers_table.transfer_id
                )
            ),
        )
        sql, args = tuple(sql_delete)
        self.client.execute(sql)

        if len(transfers) > 0:
            # batch insert of transfers
            transfer_rows = []
            account_transfer_rows = []
            for transfer in transfers:
                transfer_rows.append(get_fields_from_transfer(transfer))
                account_transfer_rows.append(
                    {"account_id": address, "transfer_id": transfer.id}
                )

            self.client.insert_many(TABLE_NAME, transfer_rows)
            self.client.insert_many(
                ACCOUNTS_TRANSFERS_TABLE_NAME, account_transfer_rows
            )

    def delete(self, transfer_id: str) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            TransfersRepositoryInterface.delete.__doc__
        )
        self.client.delete(TABLE_NAME, id=transfer_id)

    def delete_all(self) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            TransfersRepositoryInterface.delete_all.__doc__
        )
        self.client.clear(TABLE_NAME)

    def count(self, address) -> int:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            TransfersRepositoryInterface.count.__doc__
        )
        row = self.client.select_one(
            f"SELECT count(transfer_id) FROM {ACCOUNTS_TRANSFERS_TABLE_NAME} WHERE account_id=?",
            (address,),
        )
        if row is None:
            return 0

        return row[0]


def get_fields_from_transfer(transfer: Transfer) -> dict:
    """
    Return a dict of supported fields with normalized value

    :param transfer: Transfer instance
    :return:
    """
    fields = {}
    for (key, value) in transfer.__dict__.items():
        if key.startswith("_"):
            continue
        fields[key] = value

    return fields


def get_transfer_from_row(row: tuple) -> Transfer:
    """
    Return a Transfer instance from a result set row

    :param row: Result set row
    :return:
    """
    values: List[Any] = []
    count = 0
    for value in row:
        if count == 8:
            # python datetime does not handle timezone until 3.11
            values.append(parser.parse(value))
        else:
            values.append(value)
        count += 1

    return Transfer(*values)  # pylint: disable=no-value-for-parameter



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/windows/account_rename.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AccountRenameDialog(object):
    def setupUi(self, AccountRenameDialog):
        AccountRenameDialog.setObjectName("AccountRenameDialog")
        AccountRenameDialog.resize(707, 107)
        self.gridLayout = QtWidgets.QGridLayout(AccountRenameDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.buttonBox = QtWidgets.QDialogButtonBox(AccountRenameDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(4)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.buttonBox.sizePolicy().hasHeightForWidth())
        self.buttonBox.setSizePolicy(sizePolicy)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setCenterButtons(True)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 2, 0, 1, 4)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 7, 3, 1, 1)
        self.addressLabel = QtWidgets.QLabel(AccountRenameDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.addressLabel.sizePolicy().hasHeightForWidth())
        self.addressLabel.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.addressLabel.setFont(font)
        self.addressLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.addressLabel.setObjectName("addressLabel")
        self.gridLayout.addWidget(self.addressLabel, 0, 0, 1, 1)
        self.nameLabel = QtWidgets.QLabel(AccountRenameDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.nameLabel.sizePolicy().hasHeightForWidth())
        self.nameLabel.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.nameLabel.setFont(font)
        self.nameLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.nameLabel.setObjectName("nameLabel")
        self.gridLayout.addWidget(self.nameLabel, 1, 0, 1, 1)
        self.nameLineEdit = QtWidgets.QLineEdit(AccountRenameDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(3)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.nameLineEdit.sizePolicy().hasHeightForWidth())
        self.nameLineEdit.setSizePolicy(sizePolicy)
        self.nameLineEdit.setObjectName("nameLineEdit")
        self.gridLayout.addWidget(self.nameLineEdit, 1, 1, 1, 3)
        self.addressValueLabel = QtWidgets.QLabel(AccountRenameDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(3)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.addressValueLabel.sizePolicy().hasHeightForWidth())
        self.addressValueLabel.setSizePolicy(sizePolicy)
        self.addressValueLabel.setText("")
        self.addressValueLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.addressValueLabel.setObjectName("addressValueLabel")
        self.gridLayout.addWidget(self.addressValueLabel, 0, 1, 1, 3)

        self.retranslateUi(AccountRenameDialog)
        self.buttonBox.accepted.connect(AccountRenameDialog.accept) # type: ignore
        self.buttonBox.rejected.connect(AccountRenameDialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(AccountRenameDialog)

    def retranslateUi(self, AccountRenameDialog):
        #  _translate = QtCore.QCoreApplication.translate
        AccountRenameDialog.setWindowTitle(_("Rename account"))
        self.addressLabel.setText(_("Address"))
        self.nameLabel.setText(_("Name"))



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
from typing import TYPE_CHECKING, Dict, List, Optional

from scalecodec import ScaleBytes

from tikka.domains.entities.account import AccountBalance
from tikka.interfaces.adapters.network.connection import NetworkConnectionError
from tikka.interfaces.adapters.network.node.accounts import (
    NodeAccountsException,
    NodeAccountsInterface,
)
from tikka.libs.keypair import Keypair

if TYPE_CHECKING:
    from tikka.adapters.network.node.node import NetworkNode


class NodeAccounts(NodeAccountsInterface):
    """
    NodeAccounts class
    """

    def __init__(self, node: "NetworkNode") -> None:
        """
        Use NetworkNodeInterface to request/send smiths information

        :param node: NetworkNodeInterface instance
        :return:
        """
        self.node = node

    def get_balance(self, address: str) -> Optional[AccountBalance]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodeAccountsInterface.get_balance.__doc__
        )
        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeAccountsException(NetworkConnectionError())

        # system.account: FrameSystemAccountInfo
        # {
        #   nonce: 1
        #   consumers: 0
        #   providers: 1
        #   sufficients: 0
        #   data: {
        #     randomId: 0x18a4d...
        #     free: 9,799
        #     reserved: 0
        #     feeFrozen: 0
        #   }
        # }
        balance = None

        # {
        #   /// full balance owned
        #   total: number,
        #
        #   /// full available balance (include unclaimed UDs, but without existence deposit and reserved funds)
        #   transferable: number,
        #
        #   /// Unclaimed UDs
        #   unclaim_uds: number
        # }
        # fixme: should use a runtime_call() in substrate_client, see https://forum.duniter.org/t/proposition-pour-obtenir-facilement-le-solde-dun-compte-sans-indexeur/13154/5
        public_key = Keypair(ss58_address=address).public_key.hex()
        try:
            result = self.node.connection.client.rpc_request(
                "state_call", ["UniversalDividendApi_account_balances", public_key]
            )
        except Exception as exception:
            logging.exception(exception)
            return self.get_balance_obsolete(address)
        else:
            if result is not None:
                result_obj = (
                    self.node.connection.client.runtime_config.create_scale_object(
                        "(u64,u64,u64)"
                    )
                )
                result_obj.decode(ScaleBytes(result["result"]), check_remaining=True)
                balance = AccountBalance(
                    # total balance
                    result_obj.value[0],
                    # balance without existential deposit, reserved funds and unclaimed UDs
                    result_obj.value[1] - result_obj.value[2],
                    # existential deposit + reserved funds
                    result_obj.value[0] - result_obj.value[1],
                )

        return balance

    def get_balance_obsolete(self, address: str) -> Optional[AccountBalance]:
        """
        Return AccountBalance by calculating manually the unclaimed UDs

        :param address: Account address
        :return:
        """
        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeAccountsException(NetworkConnectionError())

        # compatibility with duniter 900 (gdev)
        # todo: remove as soon as gdev is upgraded with Duniter 0.10.x
        try:
            result = self.node.connection.client.query("System", "Account", [address])
        except Exception as exception:
            logging.exception(exception)
            raise NodeAccountsException(exception)

        if result is not None:
            balance = AccountBalance(
                result["data"]["free"] + result["data"]["reserved"],
                result["data"]["free"],
                result["data"]["reserved"],
            )
        else:
            return None

        try:
            result = self.node.connection.client.get_constant(
                "Balances", "ExistentialDeposit"
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeAccountsException(exception)

        balance.reserved += result

        # get unclaimed UDs amount to calculate total balance
        try:
            identity = self.node.identities.get_identity(address)
        except Exception as exception:
            logging.exception(exception)
            raise NodeAccountsException(exception)

        if identity is not None:
            try:
                result = self.node.connection.client.query(
                    "UniversalDividend", "CurrentUdIndex"
                )
            except Exception as exception:
                logging.exception(exception)
                raise NodeAccountsException(exception)

            current_index = result
            if current_index is not None:
                try:
                    result = self.node.connection.client.query(
                        "UniversalDividend", "PastReevals"
                    )
                except Exception as exception:
                    logging.exception(exception)
                    raise NodeAccountsException(exception)

                if result is not None:
                    unclaimed_uds_balance = 0
                    index = current_index
                    for reeval_index, reeval_value in reversed(result):
                        if reeval_index <= identity.first_eligible_ud:
                            count = index - identity.first_eligible_ud
                            unclaimed_uds_balance += count * reeval_value
                            break
                        else:
                            count = index - reeval_index
                            unclaimed_uds_balance += count * reeval_value
                            index = reeval_index
                    balance.total += unclaimed_uds_balance

        return balance

    def get_balances(self, addresses: List[str]) -> Dict[str, Optional[AccountBalance]]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodeAccountsInterface.get_balances.__doc__
        )
        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeAccountsException(NetworkConnectionError())

        balances: Dict[str, Optional[AccountBalance]] = {}
        for address in addresses:

            balance = None

            # fixme: should use a runtime_call() in substrate_client, see https://forum.duniter.org/t/proposition-pour-obtenir-facilement-le-solde-dun-compte-sans-indexeur/13154/5
            public_key = Keypair(ss58_address=address).public_key.hex()
            try:
                result = self.node.connection.client.rpc_request(
                    "state_call", ["UniversalDividendApi_account_balances", public_key]
                )
            except Exception as exception:
                logging.exception(exception)
                return self.get_balances_obsolete(addresses)
            else:
                if result is not None:
                    result_obj = (
                        self.node.connection.client.runtime_config.create_scale_object(
                            "(u64,u64,u64)"
                        )
                    )
                    result_obj.decode(
                        ScaleBytes(result["result"]), check_remaining=True
                    )
                    balance = AccountBalance(
                        # total balance
                        result_obj.value[0],
                        # balance without existential deposit, reserved funds and unclaimed UDs
                        result_obj.value[1] - result_obj.value[2],
                        # existential deposit + reserved funds
                        result_obj.value[0] - result_obj.value[1],
                    )
                    balances[address] = balance
                else:
                    balances[address] = None

        return balances

    def get_balances_obsolete(
        self, addresses: List[str]
    ) -> Dict[str, Optional[AccountBalance]]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodeAccountsInterface.get_balances.__doc__
        )
        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeAccountsException(NetworkConnectionError())

        try:
            result = self.node.connection.client.query(
                "UniversalDividend", "CurrentUdIndex"
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeAccountsException(exception)

        current_ud_index = result

        try:
            reevals_result = self.node.connection.client.query(
                "UniversalDividend", "PastReevals"
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeAccountsException(exception)

        if reevals_result is None:
            raise NodeAccountsException(
                "get_balances_obsolete UniversalDividend.PastReevals is None"
            )

        try:
            identities = self.node.identities.get_identities(addresses)
        except Exception as exception:
            logging.exception(exception)
            raise NodeAccountsException(exception)

        try:
            existential_deposit_result = self.node.connection.client.get_constant(
                "Balances", "ExistentialDeposit"
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeAccountsException(exception)

        storage_functions = []
        for address in addresses:
            storage_functions.append(("System", "Account", [address]))

        try:
            multi_result = self.node.connection.client.query_multi(storage_functions)
        except Exception as exception:
            logging.exception(exception)
            raise NodeAccountsException(exception)

        balances: Dict[str, Optional[AccountBalance]] = {}
        for index, value_obj in enumerate(multi_result):
            if value_obj is None:
                balances[addresses[index]] = None
            else:
                balances[addresses[index]] = AccountBalance(
                    value_obj["data"]["free"] + value_obj["data"]["reserved"],
                    value_obj["data"]["free"],
                    value_obj["data"]["reserved"],
                )
                if existential_deposit_result is not None:
                    balances[addresses[index]].reserved += existential_deposit_result  # type: ignore

                if (
                    identities[addresses[index]] is not None
                    and current_ud_index is not None
                ):
                    unclaimed_uds_balance = 0
                    current_ud_index_ = current_ud_index
                    for reeval_index, reeval_value in reversed(reevals_result):
                        if (
                            reeval_index
                            <= identities[addresses[index]].first_eligible_ud  # type: ignore
                        ):
                            count = (
                                current_ud_index_
                                - identities[addresses[index]].first_eligible_ud  # type: ignore
                            )
                            unclaimed_uds_balance += count * reeval_value
                            break
                        else:
                            count = current_ud_index_ - reeval_index
                            unclaimed_uds_balance += count * reeval_value
                            current_ud_index_ = reeval_index
                    balances[addresses[index]].total += unclaimed_uds_balance  # type: ignore

        return balances
